import json
from os.path import exists

import aiofiles

from ..exceptions import WiserExtraConfigError


class _WiserExtraConfig:
    def __init__(self, config_file: str, hub_name: str):
        self._config_file = config_file + "_" + hub_name
        self._config = {}

    async def async_load_config(self):
        if exists(self._config_file):
            async with aiofiles.open(self._config_file, mode="r") as config_file:
                try:
                    contents = await config_file.read()
                    if contents:
                        self._config = json.loads(contents)
                except (
                    OSError,
                    EOFError,
                    TypeError,
                    AttributeError,
                    json.JSONDecodeError,
                ) as ex:
                    raise WiserExtraConfigError("Error loading extra config file") from ex
        else:
            await self.async_update_config("Info", "Version", "1.0.0")
        return

    async def async_update_config(self, section: str, key: str, value):
        # If section doesn't exist
        if section not in self._config:
            self._config[section] = {}

        # If key not in section
        if key and value:
            if self._config[section].get(key):
                self._config[section][key].update(value)
            else:
                self._config[section][key] = value

        await self.async_write_config()

    async def async_remove_config(self, section: str, key: str):
        if key and self._config[section].get(key):
            del self._config[section][key]

        await self.async_write_config()

    async def async_write_config(self):
        # Write to config file
        async with aiofiles.open(self._config_file, mode="w") as config_file:
            try:
                await config_file.write(json.dumps(self._config, indent=2))
            except (
                OSError,
                EOFError,
                TypeError,
                AttributeError,
                json.JSONDecodeError,
            ) as ex:
                raise WiserExtraConfigError("Error writing to extra config file") from ex

    # @property
    def config(self, section: str = None, key: str = None):
        try:
            if section:
                if key:
                    return self._config[section][key]
                else:
                    return self._config[section]
            else:
                return self._config
        except:
            return None



from .const import TEXT_OFF, TEXT_ON, TEXT_UNKNOWN, WiserDeviceModeEnum
from .helpers.device import _WiserElectricalDevice
from .helpers.equipment import _WiserEquipment


class _WiserSmartPlug(_WiserElectricalDevice):
    """Class representing a Wiser Smart Plug device"""

    @property
    def control_source(self) -> str:
        """Get the current control source of the smart plug"""
        return self._device_type_data.get("ControlSource", TEXT_UNKNOWN)

    @property
    def delivered_power(self) -> int:
        """Get the amount of current throught the plug over time"""
        return self._device_type_data.get("CurrentSummationDelivered", None)

    @property
    def equipment_id(self) -> int:
        """Get equipment id (v2 hub)"""
        return self._device_type_data.get("EquipmentId", 0)

    @property
    def equipment(self) -> _WiserEquipment | None:
        """Get equipment data"""
        return (
            _WiserEquipment(self._device_type_data.get("EquipmentData"))
            if self._device_type_data.get("EquipmentData")
            else None
        )

    @property
    def instantaneous_power(self) -> int:
        """Get the amount of current throught the plug now"""
        return self._device_type_data.get("InstantaneousDemand", None)

    @property
    def manual_state(self) -> str:
        """Get the current manual mode setting of the smart plug"""
        return self._device_type_data.get("ManualState", TEXT_UNKNOWN)

    @property
    def is_on(self) -> bool:
        """Get if the smart plug is on"""
        return True if self.output_state == TEXT_ON else False

    @property
    def output_state(self) -> str:
        """Get plug output state"""
        return self._device_type_data.get("OutputState", TEXT_OFF)

    @property
    def scheduled_state(self) -> str:
        """Get the current scheduled state of the smart plug"""
        return self._device_type_data.get("ScheduledState", TEXT_UNKNOWN)

    async def turn_on(self) -> bool:
        """
        Turn on the smart plug
        return: boolean
        """
        result = await self._send_command({"RequestOutput": TEXT_ON})
        if result:
            self._output_state = TEXT_ON
        return result

    async def turn_off(self) -> bool:
        """
        Turn off the smart plug
        return: boolean
        """
        result = await self._send_command({"RequestOutput": TEXT_OFF})
        if result:
            self._output_state = TEXT_OFF
        return result


class _WiserSmartPlugCollection(object):
    """Class holding all wiser smart plugs"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserSmartPlug]:
        return list(self._items)

    @property
    def available_modes(self) -> list[str]:
        return [mode.value for mode in WiserDeviceModeEnum]

    @property
    def count(self) -> int:
        return len(self.all)

    # Smartplugs
    def get_by_id(self, smartplug_id: int) -> _WiserSmartPlug:
        """
        Gets a SmartPlug object from the SmartPlugs id
        param id: id of smart plug
        return: _WiserSmartPlug object
        """
        try:
            return [
                smartplug
                for smartplug in self.all
                if smartplug.id == smartplug_id
            ][0]
        except IndexError:
            return None



"""
Handles equipment data
"""

from ..const import TEXT_UNKNOWN


class _WiserEquipmentPowerInfo:
    def __init__(self, data: dict):
        self._data = data

    @property
    def current_summation_delivered(self) -> int:
        """Get delivered power"""
        return self._data.get("CurrentSummationDelivered", None)

    @property
    def current_summation_received(self) -> int:
        """Get received power"""
        return self._data.get("CurrentSummationReceived", None)

    @property
    def total_active_power(self) -> int:
        """Get total active power"""
        return self._data.get("TotalActivePower", None)

    @property
    def active_power(self) -> int:
        """Get active power"""
        return self._data.get("ActivePower", None)

    @property
    def rms_current(self) -> int:
        """Get rms current"""
        return self._data.get("RMSCurrent", None)

    @property
    def rms_voltage(self) -> int:
        """Get rms voltage"""
        return self._data.get("RMSVoltage", None)


#  Add for a notification feature Notification if the   Power is under or over a threshold
#  for  a time (periodmins)...
class _WiserEquipmentUnderPowerNotificationInfo(object):
    def __init__(self, data: dict):
        self._data = data

    @property
    def period_mins(self) -> int:
        """Get delivered power"""
        return self._data.get("PeriodMins", 0)

    @property
    def limit(self) -> int:
        """Get limit"""
        return self._data.get("Limit", 0)

    @property
    def enabled(self) -> bool:
        """Get status enable"""
        return self._data.get("Enabled", False)


class _WiserEquipmentOverPowerNotificationInfo:
    def __init__(self, equipment_instance, data: dict):
        self._data = data
        self._equipment_instance = equipment_instance

    @property
    def period_mins(self) -> int:
        """Get period in minutes"""
        return self._data.get("PeriodMins") if self._data.get("PeriodMins") else None

    @property
    def limit(self) -> int:
        """Get limit"""
        return self._data.get("Limit", 100000) if self._data.get("Limit") else None

    @property
    def enabled(self) -> bool:
        """Get status enable"""
        return self._data.get("Enabled", False) if self._data.get("Enabled") else None


class _WiserEquipment:
    """Class to hold equipment object"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def id(self) -> int:
        """Get equipment id"""
        return self._data.get("id", 0)

    @property
    def device_id(self) -> int:
        """Get device id"""
        return self._data.get("DeviceApplicationInstanceId", 0)

    @property
    def device_type(self) -> str:
        """Get device id"""
        return self._data.get("DeviceApplicationInstanceType", TEXT_UNKNOWN)

    @property
    def uuid(self) -> str:
        """Get UUID"""
        return self._data.get("UUID", TEXT_UNKNOWN)

    @property
    def controllable(self) -> bool:
        """Get if controllable"""
        return self._data.get("Controllable", False)

    @property
    def cloud_managed(self) -> bool:
        """Get if cloud managed"""
        return self._data.get("CloudManaged", False)

    @property
    def monitored(self) -> bool:
        """Get if monitored"""
        return self._data.get("Monitored", False)

    @property
    def smart_compatible(self) -> bool:
        """Get if smart compatible"""
        return self._data.get("SmartCompatible", False)

    @property
    def smart_supported(self) -> bool:
        """Get if smart supported"""
        return self._data.get("SmartSupported", False)

    @property
    def can_be_scheduled(self) -> bool:
        """Get if can be scheduled"""
        return self._data.get("CanBeScheduled", False)

    @property
    def onoff_green_schedule_supported(self) -> bool:
        """Get if OnOff green schedule supported"""
        return self._data.get("OnOffGreenScheduleSupported", False)

    @property
    def onoff_cost_schedule_supported(self) -> bool:
        """Get if OnOff cost schedule supported"""
        return self._data.get("OnOffCostScheduleSupported", False)

    @property
    def number_of_phases(self) -> str:
        """Get number of phases"""
        return self._data.get("NumberOfPhases", TEXT_UNKNOWN)

    @property
    def configured(self) -> str:
        """Get if configured"""
        return self._data.get("Configured", TEXT_UNKNOWN)

    @property
    def installation_type(self) -> str:
        """Get installation_type"""
        return self._data.get("InstallationType", TEXT_UNKNOWN)

    @property
    def equipment_family(self) -> str:
        """Get equipment family"""
        return self._data.get("EquipmentFamily", TEXT_UNKNOWN)

    @property
    def equipment_name(self) -> str:
        """Get equipment name"""
        return self._data.get("EquipmentName", TEXT_UNKNOWN)

    @property
    def product_type(self) -> str:
        """Get equipment name"""
        return self._data.get("ProductType", TEXT_UNKNOWN)

    @property
    def functional_control_mode(self) -> str:
        """Get functional control mode"""
        return self._data.get("FunctionalControlMode", TEXT_UNKNOWN)

    @property
    def current_control_mode(self) -> str:
        """Get current control mode"""
        return self._data.get("CurrentControlMode", TEXT_UNKNOWN)

    @property
    def load_state_status(self) -> str:
        """Get load state status"""
        return self._data.get("LoadStateStatus", TEXT_UNKNOWN)

    @property
    def load_state_command_optimized(self) -> str:
        """Get load state command optimized"""
        return self._data.get("LoadStateCommandOptimized", TEXT_UNKNOWN)

    @property
    def load_shedding_status(self) -> str:
        """Get load shedding status"""
        return self._data.get("LoadSheddingStatus", TEXT_UNKNOWN)

    @property
    def load_state_command_prio(self) -> str:
        """Get load state command prio"""
        return self._data.get("LoadStateCommandPrio", TEXT_UNKNOWN)

    @property
    def load_setpoint_command_prio(self) -> int:
        """Get load setpoint command prio"""
        return self._data.get("LoadSetpointCommandPrio", None)

    @property
    def pcm_mode(self) -> bool:
        """Get Pcm Mode"""
        return self._data.get("PcmMode", False)

    @property
    def pcm_supported(self) -> bool:
        """Get Pcm supported"""
        return self._data.get("PcmSupported", False)

    @property
    def pcm_priority(self) -> int:
        """Get Pcm priority"""
        return self._data.get("PcmPriority", None)

    @property
    def direction(self) -> str:
        """Get flow direction"""
        return self._data.get("Direction", TEXT_UNKNOWN)

    @property
    def operating_status(self) -> str:
        """Get the operating status"""
        return self._data.get("OperatingStatus", TEXT_UNKNOWN)

    @property
    def fault_status(self) -> str:
        """Get the fault status"""
        return self._data.get("FaultStatus", TEXT_UNKNOWN)

    @property
    def power(self) -> _WiserEquipmentPowerInfo:
        """Get the power info"""
        return _WiserEquipmentPowerInfo(self._data)

    #  Add for a notification feature Notification if the   Power is under or over a threshold
    #  for  a time (periodmins)...
    @property
    def under_power_notification(self) -> _WiserEquipmentUnderPowerNotificationInfo:
        """Get notification info"""
        return (
            _WiserEquipmentUnderPowerNotificationInfo(
                self._data.get("UnderPowerNotification")
            )
            if self._data.get("UnderPowerNotification")
            else None
        )

    @property
    def over_power_notification(self) -> _WiserEquipmentOverPowerNotificationInfo:
        """Get notification info"""
        return (
            _WiserEquipmentUnderPowerNotificationInfo(
                self._data.get("OverPowerNotification")
            )
            if self._data.get("OverPowerNotification")
            else None
        )



from .const import TEXT_UNKNOWN
from .room import _WiserRoomCollection


class _WiserHeatingChannel(object):
    """Class representing a Wiser Heating Channel"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def demand_on_off_output(self):
        """Get the demand output for the heating channel"""
        return self._data.get("DemandOnOffOutput", TEXT_UNKNOWN)

    @property
    def heating_relay_status(self) -> str:
        """Get the state of the heating channel relay"""
        return self._data.get("HeatingRelayState", TEXT_UNKNOWN)

    @property
    def id(self) -> int:
        """Get the id of the heating channel"""
        return self._data.get("id")

    @property
    def is_smart_valve_preventing_demand(self) -> bool:
        """Get if a smart valve is preventing demand for heating channel"""
        return self._data.get("IsSmartValvePreventingDemand", False)

    @property
    def name(self) -> str:
        """Get the name of the heating channel"""
        return self._data.get("Name", TEXT_UNKNOWN)

    @property
    def percentage_demand(self) -> int:
        """Get the percentage demand of the heating channel"""
        return self._data.get("PercentageDemand", 0)

    @property
    def room_ids(self) -> list | None:
        """Get a list of the room ids attached to this heating channel"""
        return self._data.get("RoomIds", None)


class _WiserHeatingChannelCollection(object):
    """Class holding all wiser heating channel objects"""

    def __init__(
        self, heating_channel_data: dict, rooms: _WiserRoomCollection
    ):
        self._heating_channels = []
        self._heating_channel_data = heating_channel_data
        self._rooms = rooms
        self._build()

    def _build(self):
        for heat_channel in self._heating_channel_data:
            self._heating_channels.append(_WiserHeatingChannel(heat_channel))

    @property
    def all(self) -> list[_WiserHeatingChannel]:
        return list(self._heating_channels)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, channel_id: int) -> _WiserHeatingChannel:
        """
        Gets a Heating Channel object from the Heating Channels id
        param id: id of heating channel
        return: _WiserHeatingChannel object
        """
        for heating_channel in self.all:
            if channel_id == heating_channel.id:
                return heating_channel
        return None

    def get_by_room_id(self, room_id: int) -> _WiserHeatingChannel:
        """
        Gets a Heating Channel object from a Room ID
        param id: id of the room
        return: _WiserHeatingChannel object
        """
        for heating_channel in self.all:
            if room_id in heating_channel.room_ids:
                return heating_channel
        return None

    def get_by_room_name(self, room_name: str) -> _WiserHeatingChannel:
        """
        Gets a Heating Channel object from a Room Name
        param id: name of the room
        return: _WiserHeatingChannel object
        """
        room = self._rooms.get_by_name(room_name)
        return self.get_by_room_id(room.id)



import inspect

from . import _LOGGER
from .const import (
    TEXT_UNKNOWN,
    WISERDEVICE,
    WiserDeviceModeEnum,
    WiserShutterAwayActionEnum,
)
from .helpers.device import _WiserElectricalDevice


class _WiserLiftMovementRange(object):
    """Data structure for min/max output range"""

    def __init__(self, shutter_instance, data: dict):
        self._shutter_instance = shutter_instance
        self._data = data

    @property
    def open_time(self) -> int:
        """Get open time value"""
        if self._data:
            return self._data.get("LiftOpenTime")
        return None

    async def set_open_time(self, time: int):
        """Set open time"""
        return await self._shutter_instance._send_command(
            {"LiftOpenTime": time, "LiftCloseTime": self.close_time}
        )

    @property
    def close_time(self) -> int:
        """Get close time value"""
        if self._data:
            return self._data.get("LiftCloseTime")
        return None

    async def set_close_time(self, time: int):
        """Set close time"""
        return await self._shutter_instance._send_command(
            {"LiftOpenTime": self.open_time, "LiftCloseTime": time}
        )

    @property
    def tilt_time(self) -> int:
        """Get tilt time value"""
        if self._data:
            return self._data.get("TiltTime")
        return None

    async def set_tilt_time(self, time: int):
        """Set tilt time"""
        return await self._shutter_instance._send_command({"TiltTime": time})

    @property
    def tilt_angle_closed(self) -> int:
        """Get tilt time value"""
        if self._data:
            return self._data.get("TiltAngleClosed", 0)
        return None

    async def set_tilt_angle_closed(self, angle: int):
        """Set tilt angle closed"""
        return await self._shutter_instance._send_command(
            {"TiltAngleClosed": angle}
        )

    @property
    def tilt_angle_open(self) -> int:
        """Get tilt angle open"""
        if self._data:
            return self._data.get("TiltAngleOpen", 0)
        return None

    async def set_tilt_angle_open(self, angle: int):
        """Set tilt angle open"""
        return await self._shutter_instance._send_command(
            {"TiltAngleOpen": angle}
        )

    @property
    def tilt_enabled(self) -> bool:
        """Get tilt enabled"""
        if self._data:
            return self._data.get("TiltEnabled")
        return None

    async def set_tilt_enabled(self, en: bool):
        """Set tilt enable"""
        return await self._shutter_instance._send_command({"TiltEnabled": en})


class _WiserShutter(_WiserElectricalDevice):
    """Class representing a Wiser Shutter device"""

    async def _send_command(self, cmd: dict, device_level: bool = False):
        """
        Send control command to the smart plug
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        if device_level:
            result = await self._wiser_rest_controller._send_command(
                WISERDEVICE.format(self.id), cmd
            )
            if result:
                self._data = result
        else:
            result = await self._wiser_rest_controller._send_command(
                self._endpoint.format(self.shutter_id), cmd
            )
            if result:
                self._device_type_data = result
        if result:
            _LOGGER.debug(
                "Wiser shutter - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def available_away_mode_actions(self):
        return [action.value for action in WiserShutterAwayActionEnum]

    @property
    def control_source(self) -> str:
        """Get the current control source of the shutter"""
        return self._device_type_data.get("ControlSource", TEXT_UNKNOWN)

    @property
    def current_lift(self) -> bool:
        """Get amount shutter is open"""
        return self._device_type_data.get("CurrentLift", 0)

    @property
    def current_tilt(self) -> int:
        """Get current tilt of shutter"""
        return self._device_type_data.get("CurrentTilt", 0)

    @property
    def drive_config(self) -> _WiserLiftMovementRange:
        """Get open and close time drive config"""
        return _WiserLiftMovementRange(
            self, self._device_type_data.get("DriveConfig")
        )

    @property
    def is_lift_position_supported(self) -> bool:
        """Get if the shutter is open"""
        return self._device_type_data.get("IsLiftPositionSupported", False)

    @property
    def is_tilt_supported(self) -> bool:
        """Get if the shutter is open"""
        return self._device_type_data.get("IsTiltSupported", False)

    @property
    def is_open(self) -> bool:
        """Get if the shutter is open"""
        return (
            True
            if self._device_type_data.get("CurrentLift", 0) == 100
            else False
        )

    @property
    def is_closed(self) -> bool:
        """Get if the shutter is closed"""
        return (
            True
            if self._device_type_data.get("CurrentLift", 0) == 0
            else False
        )

    @property
    def is_closing(self) -> bool:
        """Get if shutter is moving-opening"""
        return (
            True
            if self._device_type_data.get("LiftMovement", TEXT_UNKNOWN)
            == "Closing"
            else False
        )

    @property
    def is_opening(self) -> bool:
        """Get if shutter is moving-opening"""
        return (
            True
            if self._device_type_data.get("LiftMovement", TEXT_UNKNOWN)
            == "Opening"
            else False
        )

    @property
    def is_stopped(self) -> bool:
        """Get if shutter is not moving"""
        return (
            True
            if self._device_type_data.get("LiftMovement", TEXT_UNKNOWN)
            == "Stopped"
            else False
        )

    @property
    def is_moving(self) -> bool:
        """Get if shutter is moving"""
        return (
            True
            if self._device_type_data.get("LiftMovement", TEXT_UNKNOWN)
            != "Stopped"
            else False
        )

    @property
    def lift_movement(self) -> str:
        """Get if shutter is moving"""
        return self._device_type_data.get("LiftMovement", TEXT_UNKNOWN)

    @property
    def manual_lift(self) -> int:
        """Get shutter manual lift value"""
        return self._device_type_data.get("ManualLift", 0)

    @property
    def manual_tilt(self) -> int:
        """Get shutter manual tilt value"""
        return self._device_type_data.get("ManualTilt", 0)

    @property
    def scheduled_lift(self) -> str:
        """Get the current scheduled lift of the shutter"""
        return self._device_type_data.get("ScheduledLift", TEXT_UNKNOWN)

    @property
    def shutter_id(self) -> int:
        """Get id of shutter"""
        return self._device_type_data.get("id", 0)

    @property
    def target_lift(self) -> int:
        """Get target position of shutter"""
        return self._device_type_data.get("TargetLift", 0)

    @property
    def tilt_movement(self) -> str:
        """Get if tilt shutter is moving"""
        return self._device_type_data.get("TiltMovement", TEXT_UNKNOWN)

    @property
    def target_tilt(self) -> int:
        """Get target tilt of shutter"""
        return self._device_type_data.get("TargetTilt", 0)

    @property
    def respect_summer_comfort(self) -> bool:
        """Get if the shutter respect summer comfort"""
        return self._device_type_data.get("RespectSummerComfort", False)

    async def set_respect_summer_comfort(self, en: bool):
        """Set respect summer comfort"""
        if await self._send_command({"RespectSummerComfort": en}):
            return True

    @property
    def summer_comfort_lift(self) -> int:
        """Get the shutter summer comfort lift"""
        return self._device_type_data.get("SummerComfortLift")

    async def set_summer_comfort_lift(self, lift: int):
        if await self._send_command({"SummerComfortLift": lift}):
            self._summer_comfort_lift = lift
            return True

    @property
    def summer_comfort_tilt(self) -> int:
        """Get the shutter summer comfort tilt"""
        return self._device_type_data.get("SummerComfortTilt")

    async def set_summer_comfort_tilt(self, tilt: int):
        if await self._send_command({"SummerComfortTilt": tilt}):
            self._summer_comfort_tilt = tilt
            return True

    async def open(self, percentage: int = 100):
        """Fully open shutter"""
        if percentage >= 0 and percentage <= 100:
            return await self._send_command(
                {
                    "RequestAction": {
                        "Action": "LiftTo",
                        "Percentage": percentage,
                    }
                }
            )
        else:
            raise ValueError(
                "Shutter open percentage must be between 0 and 100"
            )

    async def close(self):
        """Fully close shutter"""
        return await self._send_command(
            {"RequestAction": {"Action": "LiftTo", "Percentage": 0}}
        )

    async def stop(self):
        """Stop shutter during movement"""
        return await self._send_command({"RequestAction": {"Action": "Stop"}})

    async def open_tilt(self, percentage: int = 100):
        """Open tilt shutter to percentage"""
        if percentage >= 0 and percentage <= 100:
            return await self._send_command(
                {
                    "RequestAction": {
                        "Action": "TiltTo",
                        "Percentage": percentage,
                    }
                }
            )
        else:
            raise ValueError(
                "Shutter tilt percentage must be between 0 and 100"
            )

    async def close_tilt(self):
        """Fully close tilt shutter"""
        return await self._send_command(
            {"RequestAction": {"Action": "TiltTo", "Percentage": 0}}
        )

    async def stop_tilt(self):
        """Stop tilt shutter during movement"""
        return await self._send_command(
            {"RequestAction": {"Action": "StopTilt"}}
        )


class _WiserShutterCollection(object):
    """Class holding all wiser heating actuators"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserShutter]:
        return list(self._items)

    @property
    def available_modes(self) -> list[str]:
        return [mode.value for mode in WiserDeviceModeEnum]

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, shutter_id: int) -> _WiserShutter:
        """
        Gets a Shutter object from the Shutters device id
        param id: device id of shutter
        return: _WiserShutter object
        """
        try:
            return [
                shutter for shutter in self.all if shutter.id == shutter_id
            ][0]
        except IndexError:
            return None

    def get_by_shutter_id(self, shutter_id: int) -> _WiserShutter:
        """
        Gets a Shutter object from the Shutters id
        param id: id of shutter
        return: _WiserShutter object
        """
        try:
            return [
                shutter
                for shutter in self.all
                if shutter.shutter_id == shutter_id
            ][0]
        except IndexError:
            return None

    def get_by_room_id(self, room_id: int) -> list[_WiserShutter]:
        """
        Gets a Shutter object from the Shutters room_id
        param id: room_id of shutter
        return: list of _WiserShutter objects
        """
        return [shutter for shutter in self.all if shutter.room_id == room_id]



from ..const import TEXT_UNKNOWN


class _WiserFirmwareUpgradeItem:
    """Data structure for upgrade info for a Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def id(self) -> int:
        "Get the id of the firmware filename"
        return self._data.get("id", 0)

    @property
    def filename(self) -> str:
        "Get the filename of the firmware file"
        return self._data.get("FirmwareFilename", TEXT_UNKNOWN)


class _WiserFirmareUpgradeInfo:
    """Data structure to hold upgrade file info for a Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data
        self._items = []
        for item in self._data:
            self._items.append(_WiserFirmwareUpgradeItem(item))

    @property
    def all(self) -> dict:
        return self._items

    def by_id(self, id) -> _WiserFirmwareUpgradeItem:
        return [item for item in self._items if item.id == id][0]



# Exception Handlers
class WiserHubConnectionError(Exception):
    pass


class WiserHubResponseError(Exception):
    """Response error mainly for bad request"""


class WiserHubAuthenticationError(Exception):
    pass


class WiserHubRESTError(Exception):
    pass


class WiserScheduleError(Exception):
    pass


class WiserScheduleInvalidTime(Exception):
    pass


class WiserScheduleInvalidSetting(Exception):
    pass


class WiserExtraConfigError(Exception):
    pass


class WiserHubNotImplementedError(Exception):
    # _LOGGER.info("Function not yet implemented")
    pass



from ..const import WISERHUBOPENTHERM
from ..rest_controller import WiserRestActionEnum, _WiserRestController
from .temp import _WiserTemperatureFunctions as tf


class _WiserOpenThermBoilerParameters:
    """Data structure for Opentherm Boiler Parameters data"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def ch_max_setpoint_read_write(self) -> bool:
        return self._data.get("maxChSetpointReadWrite", None)

    @property
    def ch_max_setpoint_transfer_enable(self) -> bool:
        return self._data.get("maxChSetpointTransferEnable", None)

    @property
    def ch_setpoint(self) -> bool:
        return tf._from_wiser_temp(self._data.get("maxChSetpoint", None), "current")

    @property
    def ch_setpoint_lower_bound(self) -> bool:
        return tf._from_wiser_temp(
            self._data.get("maxChSetpointLowerBound", None), "current"
        )

    @property
    def ch_setpoint_upper_bound(self) -> bool:
        return tf._from_wiser_temp(
            self._data.get("maxChSetpointUpperBound", None), "current"
        )

    @property
    def hw_setpoint_read_write(self) -> bool:
        return self._data.get("dhwSetpointReadWrite", None)

    @property
    def hw_setpoint_transfer_enable(self) -> bool:
        return self._data.get("dhwSetpointTransferEnable", None)

    @property
    def hw_setpoint(self) -> bool:
        return tf._from_wiser_temp(self._data.get("dhwSetpoint", None), "current")

    @property
    def hw_setpoint_lower_bound(self) -> bool:
        return tf._from_wiser_temp(
            self._data.get("dhwSetpointLowerBound", None), "current"
        )

    @property
    def hw_setpoint_upper_bound(self) -> bool:
        return tf._from_wiser_temp(
            self._data.get("dhwSetpointUpperBound", None), "current"
        )

    @property
    def json_data(self) -> dict:
        return self._data


class _WiserOpenThermOperationalData:
    """Data structure for Opentherm Boiler Parameters data"""

    def __init__(self, data):
        self._data = data

    @property
    def ch_pressure_bar(self) -> str:
        """Get ChPressureBar"""
        return self._data.get("ChPressureBar", 0) / 10

    @property
    def ch_flow_temperature(self) -> str:
        """Get Ch1FlowTemperature"""
        return tf._from_wiser_temp(
            self._data.get("Ch1FlowTemperature", None), "current"
        )

    @property
    def ch_return_temperature(self) -> str:
        """Get ChReturnTemperature"""
        return tf._from_wiser_temp(
            self._data.get("ChReturnTemperature", None), "current"
        )

    @property
    def hw_flow_rate(self) -> str:
        """Get Dhw1Temperature"""
        return self._data.get("DhwFlowRate", None)

    @property
    def hw_temperature(self) -> str:
        """Get Dhw1Temperature"""
        return tf._from_wiser_temp(self._data.get("Dhw1Temperature", None), "current")

    @property
    def relative_modulation_level(self) -> int:
        """Get RelativeModulationLevel"""
        if self._data.get("RelativeModulationLevel"):
            return int(self._data.get("RelativeModulationLevel", 0) / 10)
        return None

    @property
    def slave_status(self) -> int:
        """Get SlaveStatus"""
        return self._data.get("SlaveStatus", None)

    @property
    def json_data(self) -> dict:
        return self._data


class _WiserOpenThermExtendedDiagnostics:
    """Data structure for Opentherm extended diagnostics."""

    def __init__(self, data: dict[str, int | str]):
        self._data = data

    @property
    def unsuccessful_burner_starts(self) -> int:
        """Get unsuccessfulBurnerStarts"""
        return self._data.get("unsuccessfulBurnerStarts", 0)

    @property
    def number_flame_signal_too_low(self) -> int:
        """Get numberFlameSignalTooLow"""
        return self._data.get("numberFlameSignalTooLow", 0)

    @property
    def oem_specific_service_code(self) -> int:
        """Get oemSpecificServiceCode"""
        return self._data.get("oemSpecificServiceCode", 0)

    @property
    def successful_burner_starts(self) -> int:
        """Get successfulBurnerStarts"""
        return self._data.get("successfulBurnerStarts", 0)

    @property
    def ch_pump_starts(self) -> int:
        """Get chPumpStarts"""
        return self._data.get("chPumpStarts", 0)

    @property
    def dhw_pump_or_valve_starts(self) -> int:
        """Get dhwPumpOrValveStarts"""
        return self._data.get("dhwPumpOrValveStarts", 0)

    @property
    def burner_starts_during_dhw_mode(self) -> int:
        """Get burnerStartsDuringDhwMode"""
        return self._data.get("burnerStartsDuringDhwMode", 0)

    @property
    def burner_hours(self) -> int:
        """Get burnerHours"""
        return self._data.get("burnerHours", 0)

    @property
    def ch_pump_hours(self) -> int:
        """Get chPumpHours"""
        return self._data.get("chPumpHours", 0)


class _WiserOpentherm(object):
    """Data structure for Opentherm data"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        data: dict,
        enabled_status: str,
    ):
        self._data = data
        self._enabled_status = enabled_status
        self._wiser_rest_controller = wiser_rest_controller

    @property
    def id(self) -> int:
        """Return system id."""
        return 0

    @property
    def ch_flow_active_lower_setpoint(self) -> float:
        """Get chFlowActiveLowerSetpoint"""
        return tf._from_wiser_temp(
            self._data.get("chFlowActiveLowerSetpoint", None), "current"
        )

    @property
    def ch_flow_active_upper_setpoint(self) -> float:
        """Get chFlowActiveUpperSetpoint"""
        return tf._from_wiser_temp(
            self._data.get("chFlowActiveUpperSetpoint", None), "current"
        )

    @property
    def ch1_flow_enabled(self) -> bool:
        """Get ch1FlowEnable"""
        return self._data.get("ch1FlowEnable", False)

    @property
    def ch1_flow_setpoint(self) -> float:
        """Get ch1FlowSetpoint"""
        return tf._from_wiser_temp(self._data.get("ch1FlowSetpoint", None), "current")

    @property
    def ch2_flow_enabled(self) -> bool:
        """Get ch2FlowEnable"""
        return self._data.get("ch2FlowEnable", False)

    @property
    def ch2_flow_setpoint(self) -> float:
        """Get ch2FlowSetpoint"""
        return tf._from_wiser_temp(self._data.get("ch2FlowSetpoint", None), "current")

    @property
    def connection_status(self) -> str:
        """Get opentherm connection status"""
        return self._enabled_status

    @property
    def enabled(self) -> bool:
        """Get Enabled"""
        return self._data.get("Enabled", False)

    @property
    def hw_enabled(self) -> bool:
        """Get dhwEnable"""
        return self._data.get("dhwEnable", False)

    @property
    def hw_flow_setpoint(self) -> float:
        """Get dhwFlowSetpoint"""
        return tf._from_wiser_temp(self._data.get("dhwFlowSetpoint", None), "current")

    @property
    def operating_mode(self) -> str:
        """Get operatingMode"""
        return self._data.get("operatingMode", None)

    @property
    def operational_data(self) -> _WiserOpenThermOperationalData:
        return _WiserOpenThermOperationalData(self._data.get("operationalData", {}))

    @property
    def boiler_parameters(self) -> _WiserOpenThermBoilerParameters:
        return _WiserOpenThermBoilerParameters(
            self._data.get("preDefinedRemoteBoilerParameters", {})
        )

    @property
    def extended_diagnostics(self) -> _WiserOpenThermExtendedDiagnostics:
        return _WiserOpenThermExtendedDiagnostics(
            self._data.get("extendedDiagnostics", {})
        )

    @property
    def room_setpoint(self) -> float:
        """Get roomTemperature"""
        return tf._from_wiser_temp(self._data.get("roomSetpoint", None), "current")

    @property
    def room_temperature(self) -> float:
        """Get roomTemperature"""
        return tf._from_wiser_temp(self._data.get("roomTemperature", None), "current")

    @property
    def tracked_room_id(self) -> int:
        """Get TrackedRoomId"""
        return self._data.get("TrackedRoomId", None)

    @property
    def json_data(self) -> dict:
        return self._data

    async def set_opentherm_parameter(self, endpoint: str, cmd_data: str) -> bool:
        """Allow settign of opentherm param"""
        return await self._wiser_rest_controller._do_hub_action(
            WiserRestActionEnum.PATCH,
            f"{WISERHUBOPENTHERM}{endpoint}",
            cmd_data,
        )



import inspect

from . import _LOGGER
from .const import TEXT_UNKNOWN, WISERSYSTEM
from .rest_controller import _WiserRestController


class _WiserMoment(object):
    def __init__(
        self, wiser_rest_controller: _WiserRestController, moment_data: dict
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._moment_data = moment_data

    async def _send_command(self, cmd: dict) -> bool:
        """
        Send system control command to Wiser Hub
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        result = await self._wiser_rest_controller._send_command(
            WISERSYSTEM, cmd
        )
        if result:
            _LOGGER.debug(
                "Wiser hub - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def id(self) -> int:
        return self._moment_data.get("id", 0)

    @property
    def name(self) -> str:
        return self._moment_data.get("Name", TEXT_UNKNOWN)

    async def activate(self):
        """Activate moment"""
        return await self._send_command({"TriggerMoment": self.id})


class _WiserMomentCollection(object):
    def __init__(
        self, wiser_rest_controller: _WiserRestController, moments_data: dict
    ):
        self._moments_data = moments_data
        self._moments = []
        self._wiser_rest_controller = wiser_rest_controller
        self._build()

    def _build(self):
        for moment in self._moments_data:
            self._moments.append(
                _WiserMoment(self._wiser_rest_controller, moment)
            )

    @property
    def all(self) -> list[_WiserMoment]:
        """Return list of moments"""
        return self._moments

    @property
    def count(self) -> int:
        """Count of moments"""
        return len(self._moments)

    def get_by_id(self, moment_id: int) -> _WiserMoment:
        try:
            return [moment for moment in self.all if moment.id == moment_id][0]
        except IndexError:
            return None



from pathlib import Path, _ignore_error as pathlib_ignore_error
from typing import Union

import aiofiles.os as os
import enum


async def file_exists(file: str) -> bool:
    return await os.path.isfile(file)


def is_value_in_list(value: str, item_list: list) -> bool:
    for item in item_list:
        if value.casefold() == item.casefold():
            return True
    return False


def is_valid_level(level: int) -> bool:
    if level >= 0 and level <= 100:
        return True
    return False



"""
Handles hub capabilities
"""


class _WiserHubCapabilitiesInfo:
    """Data structure for device capabilities info for Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def all(self) -> dict:
        "Get the list of capabilities"
        return dict(self._data)

    @property
    def smartplug(self) -> bool:
        return self._data.get("SmartPlug", False)

    @property
    def itrv(self) -> bool:
        return self._data.get("ITRV", False)

    @property
    def roomstat(self) -> bool:
        return self._data.get("Roomstat", False)

    @property
    def ufh(self) -> bool:
        return self._data.get("UFH", False)

    @property
    def ufh_floor_temp_sensor(self) -> bool:
        return self._data.get("UFHFloorTempSensor", False)

    @property
    def ufh_dew_sensor(self) -> bool:
        return self._data.get("UFHDewSensor", False)

    @property
    def hact(self) -> bool:
        return self._data.get("HACT", False)

    @property
    def lact(self) -> bool:
        return self._data.get("LACT", False)

    @property
    def light(self) -> bool:
        return self._data.get("Light", False)

    @property
    def shutter(self) -> bool:
        return self._data.get("Shutter", False)

    @property
    def load_controller(self) -> bool:
        return self._data.get("LoadController", False)

    # Hub V2 additional capabilities

    @property
    def smart_socket(self) -> bool:
        return self._data.get("SmartSocket", False)

    @property
    def two_gang_lights(self) -> bool:
        return self._data.get("TwoGangLights", False)

    @property
    def fls(self) -> bool:
        return self._data.get("FLS", False)

    @property
    def boiler_interfce(self) -> bool:
        return self._data.get("BoilerInterface", False)

    @property
    def window_door_sensor(self) -> bool:
        return self._data.get("WindowDoorSensor", False)

    @property
    def motion_light_sensor(self) -> bool:
        return self._data.get("MotionLightSensor", False)

    @property
    def water_leakage_sensor(self) -> bool:
        return self._data.get("WaterLeakageSensor", False)

    @property
    def temperature_humidity_sensor(self) -> bool:
        return self._data.get("TemperatureHumiditySensor", False)

    @property
    def power_tag_e(self) -> bool:
        return self._data.get("PowerTagE", False)

    @property
    def cfmt(self) -> bool:
        return self._data.get("CFMT", False)

    @property
    def evse(self) -> bool:
        return self._data.get("EVSE", False)

    @property
    def smoke_alarm(self) -> bool:
        return self._data.get("SmokeAlarmDevice", False)

    @property
    def air_zone(self) -> bool:
        return self._data.get("Airzone", False)

    @property
    def nod_on_puck(self) -> bool:
        return self._data.get("NodOnPuck", False)

    @property
    def nod_on_16_puck(self) -> bool:
        return self._data.get("NodOn16APuck", False)

    @property
    def ev_socket(self) -> bool:
        return self._data.get("EVSocket", False)

    @property
    def fil_pilote_puck(self) -> bool:
        return self._data.get("FilPilotePuck", False)

    @property
    def iconic_devices(self) -> bool:
        return self._data.get("IconicDevices", False)


class _WiserHubAutomationCapabilities:
    """Data structure for automations features (v2 hub)"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def max_actions(self) -> int:
        return self._data.get("MaxActions", 0)

    @property
    def max_triggers(self) -> int:
        return self._data.get("MaxTriggers", 0)

    @property
    def max_constraints(self) -> int:
        return self._data.get("MaxTimeConstraints", 0)


class _WiserHubPTECapabilities:
    """Data structure for PTE features (v2 hub)"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def energy_export(self) -> bool:
        return self._data.get("EnergyExport", False)


class _WiserHubFeatureCapabilitiesInfo:
    """Data structure for feature capabilities info for Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def all(self) -> dict:
        "Get the list of capabilities"
        return dict(self._data)

    @property
    def automations(self) -> _WiserHubAutomationCapabilities:
        """Get automation capabilities"""
        return _WiserHubAutomationCapabilities(
            self._data.get("Automation", {})
        )

    @property
    def pte(self) -> _WiserHubPTECapabilities:
        """Get PTE capabilities"""
        return _WiserHubPTECapabilities(self._data.get("PTE", {}))


class _WiserClimateCapabilities(object):
    """Data structure for climate capalbilities of a room"""

    def __init__(self, room, data: dict):
        self._data = data
        self._room = room

    @property
    def heating_supported(self) -> bool:
        """Get heating supported value"""
        return self._data.get("HeatingSupported", False)

    @property
    def cooling_supported(self) -> bool:
        """Get cooling supported value"""
        return self._data.get("CoolingSupported", False)

    @property
    def minimum_heat_set_point(self) -> int:
        """Get minimum heat setpoint value"""
        return self._data.get("MinimumHeatSetpoint")

    @property
    def maximum_heat_set_point(self) -> int:
        """Get maximum heat setpoint value"""
        return self._data.get("MaximumHeatSetpoint")

    @property
    def minimum_cool_set_point(self) -> int:
        """Get minimum cool setpoint value"""
        return self._data.get("MinimumCoolSetpoint")

    @property
    def maximum_cool_set_point(self) -> int:
        """Get maximum cool setpoint value"""
        return self._data.get("MaximumCoolSetpoint")

    @property
    def setpoint_step(self) -> int:
        """Get setpoint step value"""
        return self._data.get("SetpointStep")

    @property
    def ambient_temperature(self) -> bool:
        """Get ambient temperature value"""
        return self._data.get("AmbientTemperature")

    @property
    def temperature_control(self) -> bool:
        """Get temperature control value"""
        return self._data.get("TemperatureControl")

    @property
    def open_window_detection(self) -> bool:
        """Get open window detection value"""
        return self._data.get("OpenWindowDetection", False)

    @property
    def hydronic_channel_selection(self) -> bool:
        """Get hydronic channel selection value"""
        if self._data:
            return self._data.get("HydronicChannelSelection", False)
        return None

    @property
    def on_off_supported(self) -> bool:
        """Get on off supported value"""
        if self._data:
            return self._data.get("OnOffSupported", False)
        return None



import inspect

from . import _LOGGER
from .const import (
    TEMP_MAXIMUM,
    TEMP_OFF,
    TEXT_UNKNOWN,
    WISERDEVICE,
    WISERUFHCONTROLLER,
)
from .helpers.device import _WiserDevice
from .helpers.temp import _WiserTemperatureFunctions as tf
from .rest_controller import _WiserRestController


class _WiserUFHRelay(object):
    def __init__(self, relay_data: dict):
        self.demand_percentage = relay_data.get("DemandPercentage", 0)
        self.polarity = relay_data.get("Polarity", False)
        self.id = relay_data.get("id", 0)


class _WiserUFHController(_WiserDevice):
    """Class representing a Wiser Heating Actuator device"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        endpoint: str,
        data: dict,
        device_type_data: dict,
        schedule: dict = None,
    ):
        super().__init__(
            wiser_rest_controller, endpoint, data, device_type_data
        )
        self._relays = []
        self.build_relay_collection(self._device_type_data.get("Relays", []))

    def build_relay_collection(self, relays: dict):
        """Build collection of relays"""
        for relay in relays:
            self._relays.append(_WiserUFHRelay(relay))

    def _send_command(self, cmd: dict, device_level: bool = False):
        """
        Send control command to the heating actuator
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        if device_level:
            result = self._wiser_rest_controller._send_command(
                WISERDEVICE.format(self.id), cmd
            )
        else:
            result = self._wiser_rest_controller._send_command(
                WISERUFHCONTROLLER.format(self.id), cmd
            )
        if result:
            _LOGGER.debug(
                "Wiser UFH Controller - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
        return result

    @property
    def current_temperature(self) -> float:
        """Get the current temperature measured by the smart valve"""
        return tf._from_wiser_temp(
            self._device_type_data.get("MeasuredTemperature"), "current"
        )

    @property
    def dew_detected(self) -> bool:
        """Get if dew detected"""
        return self._device_type_data.get("DewDetected", None)

    @property
    def interlock_active(self) -> bool:
        """Get if interlock active"""
        return self._device_type_data.get("InterlockActive", None)

    @property
    def is_full_strip(self) -> bool:
        """Get if full strip"""
        return self._device_type_data.get("IsFullStrip", None)

    @property
    def max_floor_temperature(self) -> int:
        """Get the max heat floor temperature"""
        return self._device_type_data.get(
            "MaxHeatFloorTemperature", TEMP_MAXIMUM
        )

    @property
    def min_floor_temperature(self) -> int:
        """Get the min heat floor temperature"""
        return self._device_type_data.get("MinHeatFloorTemperature", TEMP_OFF)

    @property
    def name(self) -> str:
        """Get name of UFH controller"""
        return self._device_type_data.get("Name", TEXT_UNKNOWN)

    @property
    def output_type(self) -> str:
        """Get output type"""
        return self._device_type_data.get("OutputType", TEXT_UNKNOWN)

    @property
    def relays(self) -> list:
        return self._relays


class _WiserUFHControllerCollection(object):
    """Class holding all wiser heating actuators"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserUFHController]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, ufh_controller_id: int) -> _WiserUFHController:
        """
        Gets a Heating Actuator object from the Heating Actuators id
        param id: id of smart valve
        return: _WiserSmartValve object
        """
        try:
            return [
                ufh_controller
                for ufh_controller in self.all
                if ufh_controller.id == ufh_controller_id
            ][0]
        except IndexError:
            return None



"""Helper functions for time."""

from datetime import datetime

from aioWiserHeatAPI.const import WEEKDAYS, WEEKENDS


def _format_output(suntimes: list[str]) -> dict[str, str]:
    output = {}
    today = datetime.today().weekday()
    days = WEEKDAYS + WEEKENDS
    for i, suntime in enumerate(suntimes[(6 - today) :]):
        # Add day of week and formatted time
        if i < 7:
            output.update({days[(i) % 7]: _format_time(suntime)})
    return output


def _format_time(time: int) -> str:
    """Format int time format to str."""
    time = str(time).rjust(4, "0")
    return time[:2] + ":" + time[2:]


def sunrise_times(times) -> dict[str, str]:
    """Return dict of sunrise times."""
    return _format_output(times)


def sunset_times(times) -> dict[str, str]:
    """Return dict of sunset times."""
    return _format_output(times)



"""Module to manage rest commands of Wiser hub"""

import asyncio
import enum
import json
import re
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

import aiohttp
import aiohttp.client_exceptions
import aiohttp.http_exceptions

from . import _LOGGER
from .const import (
    DEFAULT_BOOST_DELTA,
    REST_RETRY_BACKOFF,
    REST_TIMEOUT,
    WISERHUBDOMAIN,
    WISERHUBSCHEDULES,
    WiserUnitsEnum,
)
from .exceptions import (
    WiserExtraConfigError,
    WiserHubAuthenticationError,
    WiserHubConnectionError,
    WiserHubResponseError,
    WiserHubRESTError,
)
from .helpers.extra_config import _WiserExtraConfig


@dataclass
class WiserAPIParams:
    """Class to hold default values"""

    passive_mode_increment: str = 0.5
    stored_manual_target_temperature_alt_source: str = "current"
    boost_temp_delta: int = DEFAULT_BOOST_DELTA
    hw_climate_mode: bool = False


# Connection info class
class _WiserConnectionInfo(object):
    def __init__(self):
        self.host: str | None = None
        self.secret: str | None = None
        self.port: int | None = None
        self.units = WiserUnitsEnum.metric
        self.extra_config_file: str | None = None
        self.enable_automations: bool = False


# Enums
class WiserRestActionEnum(enum.Enum):
    """Enumeration of http methods"""

    GET = "get"
    POST = "post"
    PATCH = "patch"
    DELETE = "delete"


class _WiserRestController(object):
    """
    Class to handle getting data from and sending commands to a wiser hub
    """

    def __init__(
        self,
        timeout: Optional[float] = REST_TIMEOUT,
        wiser_connection_info: Optional[_WiserConnectionInfo] = None,
    ):
        self._wiser_connection_info = wiser_connection_info
        self._api_parameters = WiserAPIParams()
        self._timeout = aiohttp.ClientTimeout(total=timeout)
        self._hub_name = None
        self._extra_config_file = None
        self._extra_config: _WiserExtraConfig = None

        self._last_exception = None

    def remove_control_characters(self, data: str):
        """Remove control charactwers from string."""
        return re.sub(r"[\x00-\x1f]", "", data)

    async def _do_hub_action(
        self,
        action: WiserRestActionEnum,
        url: str,
        data: dict = None,
        raise_for_endpoint_error: bool = True,
    ):
        """Function to retry on response errors due to inconsistant isues reading from the hub."""
        http_version = aiohttp.HttpVersion11
        self._last_exception = None

        for i in range(5):
            if i > 0:
                await asyncio.sleep(REST_RETRY_BACKOFF[i])
            try:
                start_time = datetime.now()
                _LOGGER.debug(
                    "URL: %s, Attempt: %i, Http: %s",
                    url.format(
                        self._wiser_connection_info.host,
                        self._wiser_connection_info.port,
                    ),
                    i + 1,
                    http_version,
                )
                response = await self._execute_request(
                    action,
                    url,
                    data,
                    raise_for_endpoint_error,
                    http_version,
                )
                _LOGGER.debug(
                    "Request successful and took %ss",
                    (datetime.now() - start_time).total_seconds(),
                )

            except WiserHubRESTError as ex:
                # if json error try http1.1
                _LOGGER.debug("%s. Retrying in %.1fs", ex, REST_RETRY_BACKOFF[i])
                self._last_exception = ex
                http_version = aiohttp.HttpVersion11
            except WiserHubResponseError as ex:
                # If response error try http1.0
                _LOGGER.debug("%s. Retrying in %.1fs", ex, REST_RETRY_BACKOFF[i])
                self._last_exception = ex
                http_version = aiohttp.HttpVersion10
            except WiserHubConnectionError as ex:
                # Connection timed out or failed.  Try it again
                _LOGGER.debug("%s. Retrying in %.1fs", ex, REST_RETRY_BACKOFF[i])
                self._last_exception = ex
            except Exception as ex:
                # Unknown error
                _LOGGER.debug("%s. Retrying in %.1fs", ex, REST_RETRY_BACKOFF[i])
                self._last_exception = ex
            else:
                return response

        raise WiserHubConnectionError(self._last_exception)

    async def _execute_request(
        self,
        action: WiserRestActionEnum,
        url: str,
        data: dict | None = None,
        raise_for_endpoint_error: bool = True,
        http_version=aiohttp.HttpVersion11,
    ):
        """Send request to hub and raise errors if fails.

        param url: url of hub rest api endpoint
        param patchData: json object containing command and values to set
        return: boolean
        """

        url = url.format(
            self._wiser_connection_info.host,
            self._wiser_connection_info.port,
        )

        kwargs = {}
        kwargs["headers"] = {
            "SECRET": self._wiser_connection_info.secret,
            "Accept-Encoding": "gzip, deflate, br",
            "Accept": "*/*",
            "User-Agent": "HomeAssistant",
        }

        # if data is not None:
        if data:
            kwargs["json"] = data
            kwargs["headers"].update(
                {
                    "Content-Type": "application/json;charset=UTF-8",
                }
            )
        if self._timeout is not None:
            kwargs["timeout"] = self._timeout

        try:
            async with aiohttp.ClientSession(
                version=http_version,
            ) as session:
                async with session.request(action.value, url, **kwargs) as response:
                    await asyncio.sleep(0)
                    if not response.ok:
                        self._process_nok_response(
                            response, url, data, raise_for_endpoint_error
                        )
                    else:
                        content = await response.read()
                        if len(content) > 0:
                            response = content.decode("utf-8", "ignore")
                            try:
                                return json.loads(
                                    self.remove_control_characters(response),
                                )
                            except json.decoder.JSONDecodeError as ex:
                                raise WiserHubRESTError(
                                    f"""JSON decoding error from {url}. Error is - {ex}.
                                    Data is - {content}""",
                                ) from ex
                        else:
                            return {}
                    return {}

        except asyncio.TimeoutError as ex:
            raise WiserHubConnectionError(
                f"Connection timeout trying to communicate with Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url}"
            ) from ex
        except ConnectionResetError as ex:
            raise WiserHubConnectionError(
                f"Connection was reset by the hub during communication "
                f"{self._wiser_connection_info.host} for url {url}.  Error is {ex}"
            ) from ex
        except aiohttp.ClientResponseError as ex:
            raise WiserHubResponseError(
                f"Response error trying to communicate with Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url}.  Error is {ex}"
            ) from ex
        except aiohttp.ClientConnectorError as ex:
            raise WiserHubConnectionError(
                f"Connection error trying to communicate with Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url}.  Error is {ex}"
            ) from ex
        except (
            aiohttp.client_exceptions.ClientPayloadError,
            aiohttp.http_exceptions.TransferEncodingError,
        ) as ex:
            raise WiserHubConnectionError(
                f"Response payload incomplete error trying to communicate with Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url}.  Error is {ex}"
            ) from ex

    def _process_nok_response(
        self,
        response: aiohttp.ClientResponse,
        url: str,
        data: dict,
        raise_for_endpoint_error: bool = True,
    ):
        if response.status == 400:
            _LOGGER.debug("400 error")
        if response.status == 401:
            raise WiserHubAuthenticationError(
                f"Error authenticating to Wiser Hub "
                f"{self._wiser_connection_info.host}. Check your secret key"
            )
        elif response.status == 404 and raise_for_endpoint_error:
            raise WiserHubRESTError(
                f"Rest endpoint not found on Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url}"
            )
        elif response.status == 408:
            raise WiserHubConnectionError(
                f"Connection timed out trying to communicate with Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url}"
            )
        elif raise_for_endpoint_error:
            raise WiserHubRESTError(
                f"Unknown error communicating with Wiser Hub "
                f"{self._wiser_connection_info.host} for url {url} with data {data}. "
                f"Error code is: {response.status}"
            )

    async def get_hub_data(self, url: str, raise_for_endpoint_error: bool = True):
        """Get data from hub"""
        return await self._do_hub_action(
            WiserRestActionEnum.GET,
            url,
            raise_for_endpoint_error=raise_for_endpoint_error,
        )

    async def get_extra_config_data(self):
        # Load extra config file

        if self._extra_config_file:
            try:
                self._extra_config = _WiserExtraConfig(
                    self._extra_config_file, self._hub_name.lower()
                )
                await self._extra_config.async_load_config()
            except WiserExtraConfigError:
                _LOGGER.error(
                    "Your config file is corrupted and needs to be fixed "
                    "to maintain all the functionality of this integration."
                )
                self._extra_config = None

    async def _send_command(
        self,
        url: str,
        command_data: dict,
        method: WiserRestActionEnum = WiserRestActionEnum.PATCH,
    ):
        """
        Send control command to hub and raise errors if fails
        param url: url of hub rest api endpoint
        param patchData: json object containing command and values to set
        return: boolean
        """
        url = WISERHUBDOMAIN + url
        _LOGGER.debug(
            "Sending command to url: %s with parameters %s", url, command_data
        )

        return await self._do_hub_action(method, url, command_data)

    async def _do_schedule_action(
        self, action: WiserRestActionEnum, url: str, schedule_data: dict = None
    ):
        """
        Perform schedule action to hub and raise errors if fails
        param url: url of hub rest api endpoint
        param patchData: json object containing schedule values to set
        return: boolean
        """
        url = WISERHUBSCHEDULES + url
        _LOGGER.debug(
            "Actioning schedule to url: %s with action %s and data %s",
            url,
            action.value,
            schedule_data,
        )

        return await self._do_hub_action(action, url, schedule_data)

    async def _send_schedule_command(
        self,
        action: str,
        schedule_data: dict,
        schedule_id: int = 0,
        schedule_type: str = None,
    ) -> bool:
        """
        Send schedule data to Wiser Hub
        param schedule_data: json schedule data
        param id: schedule id
        return: boolen - true = success, false = failed
        """
        if action == "UPDATE":
            result = await self._do_schedule_action(
                WiserRestActionEnum.PATCH,
                f"{schedule_type}/{schedule_id}",
                schedule_data,
            )

        elif action == "CREATE":
            result = await self._do_schedule_action(
                WiserRestActionEnum.POST,
                "Assign",
                schedule_data,
            )

        elif action == "ASSIGN":
            result = await self._do_schedule_action(
                WiserRestActionEnum.PATCH,
                "Assign",
                schedule_data,
            )

        elif action == "DELETE":
            result = await self._do_schedule_action(
                WiserRestActionEnum.DELETE,
                f"{schedule_type}/{schedule_id}",
                schedule_data,
            )
        else:
            result = None
        return result



"""
Module to manage all devices
"""

import enum

from .binary_sensor import (
    _WiserBinarySensorCollection,
    _WiserWindowDoorSensor,
)
from .boiler_interface import _WiserBoilerInterface, _WiserBoilerInterfaceCollection
from .const import (
    TEXT_UNKNOWN,
    WISERBINARYSENSOR,
    WISERBOILERINTERFACE,
    WISERHEATINGACTUATOR,
    WISERLIGHT,
    WISERPOWERTAGENERGY,
    WISERROOMSTAT,
    WISERSHUTTER,
    WISERSMARTPLUG,
    WISERSMARTVALVE,
    WISERSMOKEALARM,
    WISERUFHCONTROLLER,
)
from .heating_actuator import (
    _WiserHeatingActuator,
    _WiserHeatingActuatorCollection,
)
from .helpers.device import _WiserDevice
from .light import _WiserDimmableLight, _WiserLight, _WiserLightCollection
from .pte import _WiserPowerTagEnergy, _WiserPowerTagEnergyCollection
from .rest_controller import _WiserRestController
from .roomstat import _WiserRoomStat, _WiserRoomStatCollection
from .schedule import WiserScheduleTypeEnum, _WiserScheduleCollection
from .shutter import _WiserShutter, _WiserShutterCollection
from .smartplug import _WiserSmartPlug, _WiserSmartPlugCollection
from .smartvalve import _WiserSmartValve, _WiserSmartValveCollection
from .smokealarm import _WiserSmokeAlarm, _WiserSmokeAlarmCollection
from .ufh import _WiserUFHController, _WiserUFHControllerCollection


class _WiserDeviceTypeEnum(enum.Enum):
    iTRV = "SmartValve"
    RoomStat = "RoomStat"
    SmartPlug = "SmartPlug"
    HeatingActuator = "HeatingActuator"
    UnderFloorHeating = "UnderFloorHeating"
    Shutter = "Shutter"
    OnOffLight = "Light"
    DimmableLight = "Light"
    PowerTagE = "PTE"
    SmokeAlarmDevice = "SmokeAlarmDevice"
    WindowDoorSensor = "BinarySensor"
    BoilerInterface = "BoilerInterface"
    CFMT = "HeatingActuator"


PRODUCT_TYPE_CONFIG = {
    "iTRV": {
        "class": _WiserSmartValve,
        "collection": _WiserSmartValveCollection,
        "endpoint": WISERSMARTVALVE,
        "heating": True,
    },
    "RoomStat": {
        "class": _WiserRoomStat,
        "collection": _WiserRoomStatCollection,
        "endpoint": WISERROOMSTAT,
        "heating": True,
    },
    "HeatingActuator": {
        "class": _WiserHeatingActuator,
        "collection": _WiserHeatingActuatorCollection,
        "endpoint": WISERHEATINGACTUATOR,
        "heating": True,
        "has_v2_equipment": True,
    },
    "UnderFloorHeating": {
        "class": _WiserUFHController,
        "collection": _WiserUFHControllerCollection,
        "endpoint": WISERUFHCONTROLLER,
        "heating": True,
    },
    "SmartPlug": {
        "class": _WiserSmartPlug,
        "collection": _WiserSmartPlugCollection,
        "endpoint": WISERSMARTPLUG,
        "schedule_type": WiserScheduleTypeEnum.onoff,
        "has_v2_equipment": True,
    },
    "Shutter": {
        "class": _WiserShutter,
        "collection": _WiserShutterCollection,
        "endpoint": WISERSHUTTER,
        "device_id_field": "DeviceId",
        "schedule_type": WiserScheduleTypeEnum.level,
    },
    "OnOffLight": {
        "class": _WiserLight,
        "collection": _WiserLightCollection,
        "endpoint": WISERLIGHT,
        "device_id_field": "DeviceId",
        "schedule_type": WiserScheduleTypeEnum.level,
    },
    "DimmableLight": {
        "class": _WiserDimmableLight,
        "collection": _WiserLightCollection,
        "endpoint": WISERLIGHT,
        "device_id_field": "DeviceId",
        "schedule_type": WiserScheduleTypeEnum.level,
    },
    "PowerTagE": {
        "class": _WiserPowerTagEnergy,
        "collection": _WiserPowerTagEnergyCollection,
        "endpoint": WISERPOWERTAGENERGY,
        "device_id_field": "DeviceId",
        "has_v2_equipment": True,
    },
    "SmokeAlarmDevice": {
        "class": _WiserSmokeAlarm,
        "collection": _WiserSmokeAlarmCollection,
        "endpoint": WISERSMOKEALARM,
        "device_id_field": "DeviceId",
    },
    "WindowDoorSensor": {
        "class": _WiserWindowDoorSensor,
        "collection": _WiserBinarySensorCollection,
        "endpoint": WISERBINARYSENSOR,
        "device_id_field": "DeviceId",
    },
    "BoilerInterface": {
        "class": _WiserBoilerInterface,
        "collection": _WiserBoilerInterfaceCollection,
        "endpoint": WISERBOILERINTERFACE,
        "device_id_field": "DeviceId",
    },
    "CFMT": {
        "class": _WiserHeatingActuator,
        "collection": _WiserHeatingActuatorCollection,
        "endpoint": WISERHEATINGACTUATOR,
        "heating": True,
        "has_v2_equipment": True,
    },
}


class _WiserDeviceCollection:
    """Class holding all wiser devices"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        domain_data: dict,
        schedules: _WiserScheduleCollection,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._device_data = domain_data.get("Device", {})
        self._domain_data = domain_data
        self._equipment_data = domain_data.get("Equipment", {})  # Supported by v2 hub
        self._schedules = schedules
        self._device_collection = {}

        self._build()

    def _get_equipment_data(self, equipment_id: int) -> dict:
        """Get equipment data"""
        equipment_data = [
            equipment_data
            for equipment_data in self._equipment_data
            if equipment_data.get("id") == equipment_id
        ]
        return equipment_data[0] if equipment_data else None

    def _build(self):
        """Build collection of devices by type"""

        # Instantiate collection classes
        for device_type in _WiserDeviceTypeEnum:
            if (
                device_type.value not in self._device_collection
                and device_type.name in PRODUCT_TYPE_CONFIG
            ):
                self._device_collection[device_type.value] = PRODUCT_TYPE_CONFIG[
                    device_type.name
                ].get("collection")()

        # Iterate device data for all known device types
        if self._device_data:
            for device in self._device_data:
                device_type = device.get("ProductType", TEXT_UNKNOWN)
                if device_type in PRODUCT_TYPE_CONFIG:
                    device_config = PRODUCT_TYPE_CONFIG[device_type]
                    device_info = [
                        device_info
                        for device_info in self._domain_data.get(
                            _WiserDeviceTypeEnum[device_type].value
                        )
                        if device_info.get(device_config.get("device_id_field", "id"))
                        == device.get("id")
                    ]

                    # Get class to create and collection to append for device
                    device_class = PRODUCT_TYPE_CONFIG[device_type].get("class")

                    # If heating device add room id
                    if device_config.get("heating"):
                        if not device.get("RoomId"):
                            device_info[0]["RoomId"] = self._get_temp_device_room_id(
                                self._domain_data, device.get("id")
                            )

                    # If schedule device add schedule
                    if device_config.get("schedule_type"):
                        device_schedule = [
                            schedule
                            for schedule in self._schedules.get_by_type(
                                device_config.get("schedule_type")
                            )
                            if schedule.id == device_info[0].get("ScheduleId")
                        ]
                    else:
                        device_schedule = None

                    # If has equipment entry, add equipment to device info data
                    if device_config.get("has_v2_equipment"):
                        device_info[0]["EquipmentData"] = self._get_equipment_data(
                            device_info[0].get("EquipmentId")
                        )

                    # Add device to collection
                    self._device_collection[
                        _WiserDeviceTypeEnum[device_type].value
                    ]._items.append(
                        device_class(
                            self._wiser_rest_controller,
                            device_config.get("endpoint"),
                            device,
                            device_info[0],
                            device_schedule[0] if device_schedule else None,
                        )
                    )

    def _get_temp_device_room_id(self, domain_data: dict, device_id: int) -> int:
        rooms = domain_data.get("Room")
        for room in rooms:
            room_device_list = []
            room_device_list.extend(room.get("SmartValveIds", []))
            room_device_list.extend(room.get("HeatingActuatorIds", []))
            room_device_list.append(room.get("RoomStatId"))
            room_device_list.append(room.get("UnderFloorHeatingId"))
            if device_id in room_device_list:
                return room.get("id")
        return 0

    @property
    def all(self) -> list[_WiserDevice]:
        """Return all devices"""
        items = []
        for key in self._device_collection:
            items.extend(self._device_collection[key].all)
        return items

    @property
    def count(self) -> int:
        """Return count of devices"""
        return len(self.all)

    @property
    def heating_actuators(self) -> _WiserHeatingActuatorCollection:
        """Return all heating actuators"""
        return self._device_collection["HeatingActuator"]

    @property
    def lights(self) -> _WiserLightCollection:
        """Return all lights"""
        return self._device_collection["Light"]

    @property
    def power_tags(self) -> _WiserPowerTagEnergyCollection:
        """Return all power tags"""
        return self._device_collection["PTE"]

    @property
    def roomstats(self) -> _WiserRoomStatCollection:
        """Return all roomstats"""
        return self._device_collection["RoomStat"]

    @property
    def shutters(self) -> _WiserShutterCollection:
        """Return all shutters"""
        return self._device_collection["Shutter"]

    @property
    def smartplugs(self) -> _WiserSmartPlugCollection:
        """Return all smart plugs"""
        return self._device_collection["SmartPlug"]

    @property
    def smartvalves(self) -> _WiserSmartValveCollection:
        """Return all smart valves (iTRVs)"""
        return self._device_collection["SmartValve"]

    @property
    def smokealarms(self) -> _WiserSmokeAlarmCollection:
        """Return all smoke alarms"""
        return self._device_collection["SmokeAlarmDevice"]

    @property
    def ufh_controllers(self) -> _WiserUFHControllerCollection:
        """Return all UFH controllers"""
        return self._device_collection["UnderFloorHeating"]

    @property
    def binary_sensor(self):
        """Return all binary sensors"""
        return self._device_collection["BinarySensor"]

    def get_by_id(self, device_id: int) -> _WiserDevice:
        """
        Gets a device object from the devices id
        param id: id of device
        return: Device type class
        """
        try:
            return [device for device in self.all if device.id == device_id][0]
        except IndexError:
            return None

    def get_by_room_id(self, room_id: int) -> list[_WiserDevice]:
        """
        Gets a list of devices belonging to the room id
        param room_id: the id of the room
        return: Device type class
        """
        try:
            return [device for device in self.all if device.room_id == room_id]
        except IndexError:
            return None

    def get_by_node_id(self, node_id: int) -> _WiserDevice:
        """
        Gets a device object from the devices zigbee node id
        param node_id: zigbee node id of device
        return: Device type class
        """
        try:
            return [device for device in self.all if device.node_id == node_id][0]
        except IndexError:
            return None

    def get_by_serial_number(self, serial_number: str) -> _WiserDevice:
        """
        Gets a device object from the devices serial number
        param node_id: serial number of device
        return: Device type class
        """
        try:
            return [
                device for device in self.all if device.serial_number == serial_number
            ][0]
        except IndexError:
            return None

    def get_by_parent_node_id(self, node_id: int) -> list[_WiserDevice]:
        """
        Gets a list of device from the devices zigbee parent node id
        param node_id: zigbee parent node id of device
        return: Device type class
        """
        try:
            return [device for device in self.all if device.parent_node_id == node_id]
        except IndexError:
            return None



"""
Handles power tag energy devices
"""

from .const import TEXT_UNABLE, TEXT_UNKNOWN
from .helpers.device import _WiserDevice
from .helpers.equipment import _WiserEquipment


class _WiserPowerTagEnergy(_WiserDevice):
    """Class representing a Wiser Power Tag Energy device"""

    @property
    def delivered_power(self) -> int:
        """Get current power of device"""
        return self.equipment.power.current_summation_delivered

    @property
    def energy_export(self) -> str:
        """Get energy export status"""
        return self._device_type_data.get("EnergyExport", TEXT_UNABLE)

    @property
    def equipment(self) -> _WiserEquipment | None:
        """Get equipment data"""
        return (
            _WiserEquipment(self._device_type_data.get("EquipmentData"))
            if self._device_type_data.get("EquipmentData")
            else None
        )

    @property
    def grid_limit(self) -> int:
        """Get grid limit"""
        return self._device_type_data.get("GridLimit", 0)

    @property
    def grid_limit_uom(self) -> str:
        """Get grid limit uom"""
        return self._device_type_data.get("GridLimitUom", TEXT_UNKNOWN)

    @property
    def instantaneous_power(self) -> int:
        """Get current power of device"""
        return self.equipment.power.total_active_power

    @property
    def raw_total_active_power(self) -> int:
        """Get raw total active power of device"""
        return self._device_type_data.get("RawTotalActivePower", None)

    @property
    def received_power(self) -> int:
        """Get current power of device"""
        return self.equipment.power.current_summation_received

    @property
    def rfid(self) -> int:
        """Get rfid of device"""
        return self._data.get("RfId", 0)

    @property
    def self_consumption(self) -> bool:
        """Get self consumption"""
        return self._device_type_data.get("SelfConsumption", False)


class _WiserPowerTagEnergyCollection:
    """Class representing collection of Power Tag Energy devices"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserPowerTagEnergy]:
        """Return all power tags"""
        return list(self._items)

    @property
    def count(self) -> int:
        """Return number of power tags"""
        return len(self.all)

    def get_by_id(self, device_id: int) -> _WiserPowerTagEnergy:
        """
        Gets a Power Tag Energy object from the device id
        param id: id of power tag
        return: _WiserPowerTagEnergy object
        """
        try:
            return [
                power_tag
                for power_tag in self.all
                if power_tag.id == device_id
            ][0]
        except IndexError:
            return None

    def get_by_equipment_id(self, equipment_id: int) -> _WiserPowerTagEnergy:
        """
        Gets a Power Tag Energy object from the equipment id
        param id: id of power tag
        return: _WiserPowerTagEnergy object
        """
        try:
            return [
                power_tag
                for power_tag in self.all
                if power_tag.id == equipment_id
            ][0]
        except IndexError:
            return None



import asyncio
from . import _LOGGER

from typing import cast
from zeroconf import ServiceStateChange, IPVersion
from zeroconf.asyncio import AsyncServiceBrowser, AsyncZeroconf, AsyncServiceInfo

MDNS_TIMEOUT = 2


class _WiserDiscoveredHub(object):
    def __init__(self, ip: str, hostname: str, name: str):
        self._ip = ip
        self._name = name
        self._hostname = hostname

    @property
    def ip(self) -> str:
        return self._ip

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def name(self) -> str:
        return self._name


class WiserDiscovery(object):
    """
    Class to handle mDns discovery of a wiser hub on local network
    Use discover_hub() to return list of mDns responses.
    """

    def __init__(self):
        self._discovered_hubs = []
        self.mdns_timeout = MDNS_TIMEOUT

    @property
    def mdns_timeout(self):
        return self._mdns_timeout

    @mdns_timeout.setter
    def mdns_timeout(self, mdns_timeout):
        self._mdns_timeout = mdns_timeout

    def async_on_service_state_change(
        self,
        zeroconf: AsyncZeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        """
        Look for Wiser Hub in discovered services and set IP and Name in
        global vars
        """

        loop = asyncio.get_running_loop()
        loop.create_task(
            self.async_parse_state_change(zeroconf, service_type, name, state_change)
        )

    async def async_parse_state_change(
        self, zeroconf, service_type, name, state_change
    ):

        info = AsyncServiceInfo(service_type, name)

        if state_change != ServiceStateChange.Removed:
            info_success = await info.async_request(zeroconf, 3000)

            if not info_success:
                return

            if "WiserHeat" in name:
                if state_change == ServiceStateChange.Added:
                    info_success = await info.async_request(zeroconf, 3000)

                    if not info_success:
                        return

                if info:
                    addresses = [
                        "%s:%d" % (addr, cast(int, info.port))
                        for addr in info.parsed_addresses()
                    ]
                    hub = _WiserDiscoveredHub(
                        ip=addresses[0].replace(":80", ""),
                        hostname=info.server.replace(".local.", ".local").lower(),
                        name=info.server.replace(".local.", ""),
                    )
                    _LOGGER.debug(
                        "Discovered Hub {} with IP Address {}".format(
                            info.server.replace(".local.", ""),
                            addresses[0].replace(":80", ""),
                        )
                    )
                    self._discovered_hubs.append(hub)

    async def discover_hub(self, min_search_time: int = 2, max_search_time: int = 10):
        """
        Call zeroconf service browser to find Wiser hubs on the local network.
        param (optional) min_search_time: min seconds to wait for responses before returning
        param (optional) max_search_time: max seconds to wait for responses before returning
        return: list of discovered hubs
        """
        timeout = 0

        self.aioZeroConf = AsyncZeroconf(ip_version=IPVersion.V4Only)
        services = ["_http._tcp.local."]
        self.aioBrowser = AsyncServiceBrowser(
            self.aioZeroConf.zeroconf,
            services,
            handlers=[self.async_on_service_state_change],
        )

        if self.mdns_timeout > 0:
            await asyncio.sleep(self.mdns_timeout)
            await self.async_close()
        return self._discovered_hubs

    async def async_close(self) -> None:
        assert self.aioZeroConf is not None
        assert self.aioBrowser is not None
        await self.aioBrowser.async_cancel()
        await self.aioZeroConf.async_close()



import inspect
from datetime import datetime
from typing import Union

from aioWiserHeatAPI.helpers.capabilities import _WiserClimateCapabilities

from . import _LOGGER
from .const import (
    TEMP_MINIMUM,
    TEMP_OFF,
    TEXT_BOOST,
    TEXT_MANUAL,
    TEXT_OFF,
    TEXT_ON,
    TEXT_UNKNOWN,
    WISER_BOOST_DURATION,
    WISERROOM,
    WiserHeatingModeEnum,
    WiserPresetOptionsEnum,
)
from .devices import _WiserDeviceCollection
from .helpers.misc import is_value_in_list
from .helpers.temp import _WiserTemperatureFunctions as tf
from .rest_controller import WiserRestActionEnum, _WiserRestController
from .schedule import _WiserSchedule, _WiserScheduleCollection


class _WiserRoom(object):
    """Class representing a Wiser Room entity"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        room: dict,
        schedule: _WiserSchedule,
        devices: list,
        enable_automations: bool,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._data = room
        self._schedule = schedule
        self._devices = devices
        self._enable_automations = enable_automations
        self._extra_config = (
            self._wiser_rest_controller._extra_config.config(
                "Rooms", str(self.id)
            )
            if self._wiser_rest_controller._extra_config
            else None
        )
        self._mode = self._effective_heating_mode(
            self._data.get("Mode"), self.current_target_temperature
        )

        self._name = room.get("Name")
        self._include_in_summer_comfort = room.get(
            "IncludeInSummerComfort", False
        )
        self._window_detection_active = room.get(
            "WindowDetectionActive", TEXT_UNKNOWN
        )

        self._default_extra_config = {
            "passive_mode": False,
            "min": 14,
            "max": 18,
        }

        # Add device id to schedule
        if self._schedule:
            self.schedule._assignments.append(
                {"id": self.id, "name": self.name}
            )

    def _effective_heating_mode(self, mode: str, temp: float) -> str:
        if mode.casefold() == TEXT_MANUAL.casefold() and temp == TEMP_OFF:
            return WiserHeatingModeEnum.off.value
        elif mode.casefold() == TEXT_MANUAL.casefold():
            return WiserHeatingModeEnum.manual.value
        return WiserHeatingModeEnum.auto.value

    async def _send_command(
        self,
        cmd: dict,
        method: WiserRestActionEnum = WiserRestActionEnum.PATCH,
    ):
        """
        Send control command to the room
        param cmd: json command structure
        return: boolen
        """
        result = await self._wiser_rest_controller._send_command(
            WISERROOM.format(self.id), cmd, method
        )
        if result:
            _LOGGER.debug(
                "Wiser room - {} command successful - {}".format(
                    inspect.stack()[1].function, result
                )
            )
            return True
        return False

    async def _update_extra_config(self, key: str, value) -> bool:
        if self._wiser_rest_controller._extra_config:
            try:
                await self._wiser_rest_controller._extra_config.async_update_config(
                    "Rooms", str(self.id), {key: value}
                )
                return True
            except Exception:
                return False
        return False

    async def _get_extra_config(self) -> None:
        self._extra_config = (
            self._wiser_rest_controller._extra_config.config(
                "Rooms", str(self.id)
            )
            if self._wiser_rest_controller._extra_config
            else None
        )

    @property
    def is_passive_mode(self) -> bool:
        return (
            True
            if self._enable_automations
            and self.passive_mode_enabled
            and self.mode != WiserHeatingModeEnum.off.value
            and not self.is_boosted
            and not self.is_away_mode
            else False
        )

    @property
    def passive_mode_enabled(self) -> bool:
        if self._extra_config and "passive_mode" in self._extra_config:
            # Migrate from initial version of extra config
            if (
                self._extra_config["passive_mode"] == "Disabled"
                or self._extra_config["passive_mode"] is False
            ):
                return False
            else:
                return True
        return self._default_extra_config["passive_mode"]

    @property
    def passive_mode_lower_temp(self) -> float:
        if self._extra_config and "min" in self._extra_config:
            return min(self._extra_config["min"], self.passive_mode_upper_temp)
        return min(
            self._default_extra_config["min"], self.passive_mode_upper_temp
        )

    @property
    def passive_mode_upper_temp(self) -> float:
        if (
            self.is_passive_mode
            and self.mode == WiserHeatingModeEnum.auto.value
            and self.schedule
        ):
            return self.schedule.current_setting
        else:
            if self._extra_config and "max" in self._extra_config:
                return self._extra_config["max"]
            return self._default_extra_config["max"]

    async def set_passive_mode(self, enable: bool):
        if await self._update_extra_config("passive_mode", enable):
            await self._get_extra_config()
            if not self.is_away_mode:
                # Set to min of temp range when initialised
                if enable and self.mode != WiserHeatingModeEnum.off.value:
                    await self.set_target_temperature(
                        self.passive_mode_lower_temp
                    )

                # Set target temp back to last manual temp if in manual mode
                if (
                    not enable
                ) and self.mode == WiserHeatingModeEnum.manual.value:
                    await self.set_target_temperature(
                        self.stored_manual_target_temperature
                    )
        else:
            _LOGGER.error(
                "Unable to set passive mode.  This maybe caused by an issue with your extra config file."
            )

    async def set_passive_mode_lower_temp(self, temp):
        if not await self._update_extra_config("min", temp):
            _LOGGER.error(
                "Unable to set passive lower temp.  This maybe caused by an issue with your extra config file."
            )
        await self._get_extra_config()

    async def set_passive_mode_upper_temp(self, temp):
        if not await self._update_extra_config("max", temp):
            _LOGGER.error(
                "Unable to set passive mode upper temp.  This maybe caused by an issue with your extra config file."
            )
        await self._get_extra_config()

    @property
    def available_modes(self) -> list:
        """Get available heating modes"""
        return [mode.value for mode in WiserHeatingModeEnum]

    @property
    def available_presets(self) -> list:
        """Get available preset modes"""
        # Remove advance schedule if no schedule exists or in passive mode
        if self.is_passive_mode or not self.schedule:
            return [
                mode.value
                for mode in WiserPresetOptionsEnum
                if mode != WiserPresetOptionsEnum.advance_schedule
            ]
        return [mode.value for mode in WiserPresetOptionsEnum]

    @property
    def away_mode_suppressed(self) -> str:
        """Get if away mode is suppressed for room"""
        return self._data.get("AwayModeSuppressed", False)

    @property
    def boost_end_time(self) -> datetime:
        """Get boost end timestamp"""
        if self._data.get("OverrideTimeoutUnixTime", 0) == 0:
            return None
        return datetime.fromtimestamp(
            self._data.get("OverrideTimeoutUnixTime", 0)
        )

    @property
    def boost_time_remaining(self) -> datetime:
        """Get boost time remaining"""
        if self.is_boosted:
            return (self.boost_end_time - datetime.now()).total_seconds()
        else:
            return 0

    @property
    def boost_temperature_delta(self) -> float:
        return self._wiser_rest_controller._api_parameters.boost_temp_delta

    @property
    def capabilities(self) -> _WiserClimateCapabilities:
        """Get room climate capabilities"""
        if capabilities := self._data.get("ClimateCapabilities"):
            return _WiserClimateCapabilities(self, capabilities)

    @property
    def comfort_mode_score(self) -> int:
        """Get room heating comfort mode score"""
        return self._data.get("ComfortModeScore", 0)

    @property
    def control_direction(self) -> str:
        """Get room heating control direction Heat or Cool"""
        return self._data.get("ControlDirection", TEXT_UNKNOWN)

    @property
    def current_target_temperature(self) -> float:
        """Get current target temperature for the room"""
        return tf._from_wiser_temp(
            self._data.get("DisplayedSetPoint", TEMP_MINIMUM)
        )

    @property
    def current_temperature(self) -> float:
        """Get current temperature of the room"""
        return tf._from_wiser_temp(
            self._data.get("CalculatedTemperature", TEMP_MINIMUM), "current"
        )

    @property
    def current_humidity(self) -> int:
        """Get current humidity of the room if room has a roomstat"""
        for device in self.devices:
            if hasattr(device, "current_humidity"):
                return device.current_humidity
        return None

    @property
    def demand_type(self) -> str:
        """Get room heating control direction"""
        return self._data.get("DemandType", TEXT_UNKNOWN)

    @property
    def devices(self) -> list:
        """Get devices associated with the room"""
        return self._devices

    @property
    def displayed_setpoint(self) -> float:
        """Get room heating displayed setpoint"""
        return tf._from_wiser_temp(
            self._data.get("DisplayedSetPoint", TEMP_MINIMUM), "current"
        )

    @property
    def heating_actuator_ids(self) -> list:
        """Get list of heating actuator ids associated with room"""
        return sorted(self._data.get("HeatingActuatorIds", []))

    @property
    def heating_rate(self) -> str:
        """Get room heating rate"""
        return self._data.get("HeatingRate", TEXT_UNKNOWN)

    @property
    def heating_type(self) -> str:
        """Get room heating type"""
        return self._data.get("HeatingType", TEXT_UNKNOWN)

    @property
    def heating_mode(self) -> str:
        return self._effective_heating_mode(
            self._data.get("Mode"), self.current_target_temperature
        )

    @property
    def id(self) -> int:
        """Get the id of the room"""
        return self._data.get("id")

    @property
    def is_away_mode(self) -> bool:
        """Get if the room temperature is currently set by away mode"""
        return (
            True
            if "Away"
            in self._data.get(
                "SetpointOrigin",
                self._data.get("SetPointOrigin", TEXT_UNKNOWN),
            )
            else False
        )

    @property
    def is_boosted(self) -> bool:
        """Get if the room temperature is currently boosted"""
        return (
            True
            if "Boost"
            in self._data.get(
                "SetpointOrigin",
                self._data.get("SetPointOrigin", TEXT_UNKNOWN),
            )
            else False
        )

    @property
    def is_override(self) -> bool:
        """Get if the room has an override"""
        return (
            True
            if self._data.get("OverrideType", TEXT_UNKNOWN)
            not in [TEXT_UNKNOWN, "None"]
            else False
        )

    @property
    def is_heating(self) -> bool:
        """Get if the room is currently heating"""
        return (
            True
            if self._data.get("ControlOutputState", TEXT_OFF) == TEXT_ON
            else False
        )

    @property
    def manual_target_temperature(self) -> float:
        """Get current target temperature for manual mode"""
        return tf._from_wiser_temp(
            self._data.get("ManualSetPoint", TEMP_MINIMUM)
        )

    @property
    def mode(self) -> str:
        """Get or set current mode for the room (Off, Manual, Auto)"""
        return self._mode

    async def set_mode(self, mode: Union[WiserHeatingModeEnum, str]) -> bool:
        if isinstance(mode, WiserHeatingModeEnum):
            mode = mode.value

        if is_value_in_list(mode, self.available_modes):
            # Cancel any overrides on mode change
            if self.is_override:
                await self.cancel_overrides()

            if mode == WiserHeatingModeEnum.off.value:
                await self.set_manual_temperature(TEMP_OFF)
            elif mode == WiserHeatingModeEnum.manual.value:
                if await self._send_command(
                    {"Mode": WiserHeatingModeEnum.manual.value}
                ):
                    if self.current_target_temperature == TEMP_OFF:
                        await self.set_target_temperature(
                            self.stored_manual_target_temperature
                        )
            elif mode == WiserHeatingModeEnum.auto.value:
                await self._send_command(
                    {"Mode": WiserHeatingModeEnum.auto.value}
                )

            self._mode = mode
            return True
        else:
            raise ValueError(
                f"{mode} is not a valid Room mode.  Valid modes are {self.available_modes}"
            )

    @property
    def name(self) -> str:
        """Get or set the name of the room"""
        return self._name

    async def set_name(self, name: str):
        if await self._send_command({"Name": name.title()}):
            self._name = name.title()
            return True

    @property
    def number_of_heating_actuators(self) -> int:
        """Get number of heating actuators associated with room"""
        return len(self._data.get("HeatingActuatorIds", []))

    @property
    def number_of_smartvalves(self) -> int:
        """Get number of smartvalves associated with room"""
        return len(self._data.get("SmartValveIds", []))

    @property
    def override_target_temperature(self) -> float:
        """Get the override target temperature of the room"""
        return self._data.get("OverrideSetpoint", 0)

    @property
    def override_type(self) -> str:
        """Get the current override type for the room"""
        return self._data.get("OverrideType", "None")

    @property
    def percentage_demand(self) -> int:
        """Get the percentage demand of the room"""
        return self._data.get("PercentageDemand", 0)

    @property
    def preset_mode(self) -> str:
        """Get the current preset mode"""
        if self.is_boosted:
            return TEXT_BOOST
        else:
            return None

    async def set_preset(self, preset: WiserPresetOptionsEnum | str):
        """Set the preset mode"""
        if isinstance(preset, WiserPresetOptionsEnum):
            preset = preset.value

        # Is it valid preset option?
        if preset in self.available_presets:
            if preset == WiserPresetOptionsEnum.cancel_overrides.value:
                await self.cancel_overrides()
            elif preset == WiserPresetOptionsEnum.advance_schedule.value:
                await self.schedule_advance()
            elif preset.lower().startswith(TEXT_BOOST.lower()):
                # Lookup boost duration
                duration = WISER_BOOST_DURATION[preset]
                _LOGGER.debug(
                    f"Boosting by {self.boost_temperature_delta}C for {duration} mins"
                )
                await self.boost(self.boost_temperature_delta, duration)
        else:
            raise ValueError(
                f"{preset} is not a valid preset.  Valid presets are {self.available_presets}"
            )

    @property
    def roomstat_id(self) -> int:
        """Get the id of the roomstat"""
        return self._data.get("RoomStatId", None)

    @property
    def schedule(self) -> _WiserSchedule:
        """Get the schedule for the room"""
        return self._schedule

    @property
    def schedule_id(self) -> int:
        """Get the schedule id for the room"""
        return self._data.get("ScheduleId")

    @property
    def scheduled_target_temperature(self) -> float:
        """Get the scheduled target temperature for the room"""
        return tf._from_wiser_temp(
            self._data.get("ScheduledSetPoint", TEMP_MINIMUM)
        )

    @property
    def setpoint_origin(self) -> str:
        """Get the origin of the target temperature setting for the room"""
        return self._data.get(
            "SetpointOrigin", self._data.get("SetpointOrigin", TEXT_UNKNOWN)
        )

    @property
    def smartvalve_ids(self) -> list:
        """Get list of smartvalve ids associated with room"""
        return sorted(self._data.get("SmartValveIds", []))

    @property
    def stored_manual_target_temperature(self) -> float:
        if self._extra_config and "manual_temp" in self._extra_config:
            return self._extra_config["manual_temp"]
        elif (
            self._wiser_rest_controller._api_parameters.stored_manual_target_temperature_alt_source.lower()
            == "current"
        ):
            return self.current_temperature
        elif (
            self._wiser_rest_controller._api_parameters.stored_manual_target_temperature_alt_source.lower()
            == "scheduled"
            and self.schedule
        ):
            return self.scheduled_target_temperature
        return TEMP_MINIMUM

    @property
    def target_temperature_origin(self) -> str:
        """Get the origin of the target temperature setting for the room"""
        return self.setpoint_origin

    @property
    def underfloor_heating_id(self) -> int:
        """Get the id of the underfloor heating controller"""
        return self._data.get("UnderFloorHeatingId", None)

    @property
    def underfloor_heating_relay_ids(self) -> int:
        """Get the id of the underfloor heating controller relay ids"""
        return sorted(self._data.get("UfhRelayIds", []))

    # Added by LGO
    @property
    def include_in_summer_comfort(self) -> bool:
        """Get the status of participate to the summer comfort"""
        return self._data.get("IncludeInSummerComfort")

    async def set_include_in_summer_comfort(self, enabled: bool):
        """Set the status of participate to the summer comfort"""
        if await self._send_command({"IncludeInSummerComfort": enabled}):
            self._include_in_summer_comfort = enabled
            return True

    @property
    def hvac_mode(self) -> str:
        """Get the HVAC mode of the room"""
        return self._data.get("HVACMode", TEXT_UNKNOWN)

    @property
    def floor_sensor_state(self) -> str:
        """Get the state of the floor sensor"""
        return self._data.get("FloorSensorState", TEXT_UNKNOWN)

    @property
    def occupancy(self) -> str:
        """Get the occupancy of the room"""
        return self._data.get("Occupancy", TEXT_UNKNOWN)

    @property
    def occupancy_capable(self) -> bool:
        """Get the occupancy of the room"""
        return self._data.get("OccupancyCapable", False)

    @property
    def occupied_heating_set_point(self) -> int:
        """Get the setpoint when the room is occupied"""
        return tf._from_wiser_temp(
            self._data.get("OccupiedHeatingSetPoint", TEMP_MINIMUM)
        )

    @property
    def unoccupied_heating_set_point(self) -> int:
        """Get the setpoint when the room is unoccupied"""
        return tf._from_wiser_temp(
            self._data.get("UnoccupiedHeatingSetPoint", TEMP_MINIMUM)
        )

    @property
    def climate_demand_for_ui(self) -> int:
        """Get the climate demand for UI"""
        return self._data.get("ClimateDemandForUI", 0)

    # End Added by LGO

    @property
    def window_detection_active(self) -> bool:
        """Get or set if window detection is active"""
        return self._window_detection_active

    async def set_window_detection_active(self, enabled: bool):
        if await self._send_command({"WindowDetectionActive": enabled}):
            self._window_detection_active = enabled
            return True

    @property
    def window_state(self) -> str:
        """
        Get the currently detected window state for the room.
        Window detection needs to be active
        """
        return self._data.get("WindowState", TEXT_UNKNOWN)

    async def delete(self):
        """
        Delete room from wiserhub
        """
        return await self._send_command(None, WiserRestActionEnum.DELETE)

    async def boost(self, inc_temp: float, duration: int) -> bool:
        """
        Boost the target temperature of the room
        param inc_temp: increase target temperature over current temperature by 0C to 5C
        param duration: the duration to boost the room temperature in minutes
        return: boolean
        """
        if duration == 0:
            return await self.cancel_boost()
        return await self._send_command(
            {
                "RequestOverride": {
                    "Type": "Boost",
                    "DurationMinutes": duration,
                    "IncreaseSetPointBy": tf._to_wiser_temp(
                        inc_temp, "boostDelta"
                    ),
                }
            }
        )

    async def cancel_boost(self) -> bool:
        """
        Cancel the target temperature boost of the room
        return: boolean
        """
        if self.is_boosted:
            return await self.cancel_overrides()
        else:
            return True

    async def set_target_temperature(self, temp: float) -> bool:
        """
        Set the target temperature of the room to override current schedule temp or in manual mode
        param temp: the temperature to set in C
        return: boolean
        """
        # Set manual temp in stored config
        if (
            self.mode == WiserHeatingModeEnum.manual.value
            and not self.is_passive_mode
            and temp != TEMP_OFF
        ):
            await self._update_extra_config("manual_temp", temp)

        return await self._send_command(
            {
                "RequestOverride": {
                    "Type": "Manual",
                    "SetPoint": tf._to_wiser_temp(temp),
                }
            }
        )

    async def set_target_temperature_for_duration(
        self, temp: float, duration: int
    ) -> bool:
        """
        Set the target temperature of the room to override current schedule temp or in manual mode
        param temp: the temperature to set in C
        return: boolean
        """
        # Set manual temp in stored config
        if (
            self.mode == WiserHeatingModeEnum.manual.value
            and not self.is_passive_mode
        ):
            await self._update_extra_config("manual_temp", temp)

        return await self._send_command(
            {
                "RequestOverride": {
                    "Type": "Manual",
                    "DurationMinutes": duration,
                    "SetPoint": tf._to_wiser_temp(temp),
                }
            }
        )

    async def set_manual_temperature(self, temp: float) -> bool:
        """
        Set the mode to manual with target temperature for the room
        param temp: the temperature to set in C
        return: boolean
        """
        if self.mode != WiserHeatingModeEnum.manual.value:
            await self.set_mode(WiserHeatingModeEnum.manual.value)
        return await self.set_target_temperature(temp)

    async def schedule_advance(self) -> bool:
        """
        Advance room schedule to the next scheduled time and temperature setting
        return: boolean
        """
        if await self.cancel_boost():
            return await self.set_target_temperature(
                self.schedule.next.setting
            )

    async def cancel_overrides(self) -> bool:
        """
        Cancel all overrides and set room schedule to the current temperature setting for the mode
        return: boolean
        """
        return await self._send_command({"RequestOverride": {"Type": "None"}})


class _WiserRoomCollection:
    """Class holding all wiser room objects"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        room_data: dict,
        schedules: _WiserScheduleCollection,
        devices: _WiserDeviceCollection,
        enable_automations: bool,
    ):
        super().__init__()
        self._wiser_rest_controller = wiser_rest_controller
        self._room_data = room_data
        self._schedules = schedules
        self._devices = devices
        self._enable_automations = enable_automations
        self._rooms: list(_WiserRoom) = []
        self._build()

    def _build(self):
        # Add room objects
        for room in self._room_data:
            schedule = [
                schedule
                for schedule in self._schedules
                if schedule.id == room.get("ScheduleId")
            ]
            devices = self._devices.get_by_room_id(room.get("id", 0))
            self._rooms.append(
                _WiserRoom(
                    self._wiser_rest_controller,
                    room,
                    schedule[0] if len(schedule) > 0 else None,
                    devices,
                    self._enable_automations,
                )
            )

    @property
    def all(self) -> list[_WiserRoom]:
        """Returns list of room objects"""
        return self._rooms

    @property
    def count(self) -> int:
        """Number of rooms"""
        return len(self._rooms)

    async def add(self, name):
        """
        Add new room
        param name: name of room
        """
        return await self._wiser_rest_controller._send_command(
            WISERROOM, {"name": name}, WiserRestActionEnum.POST
        )

    def get_by_id(self, id: int) -> _WiserRoom:
        """
        Gets a room object for the room by id of room
        param id: the id of the room
        return: _WiserRoom object
        """
        try:
            return [room for room in self._rooms if room.id == id][0]
        except IndexError:
            return None

    def get_by_name(self, name: str) -> _WiserRoom:
        """
        Gets a room object for the room by name of room
        param name: the name of the room
        return: _WiserRoom object
        """
        try:
            return [
                room
                for room in self._rooms
                if room.name.lower() == name.lower()
            ][0]
        except IndexError:
            return None

    def get_by_schedule_id(self, schedule_id: int) -> _WiserRoom:
        """
        Gets a room object for the room a schedule id belongs to
        param schedule_id: the id of the schedule
        return: _WiserRoom object
        """
        return [
            room for room in self._rooms if room.schedule_id == schedule_id
        ][0]

    def get_by_device_id(self, device_id: int) -> _WiserRoom:
        """
        Gets a room object for the room a device id belongs to
        param device_id: the id of the device
        return: _WiserRoom object
        """
        for room in self._rooms:
            for device in room.devices:
                if device.id == device_id:
                    return room
        return None



"""
Handles binary_sensor devices
"""

from .helpers.battery import _WiserBattery
from .helpers.device import _WiserDevice


class _WiserBinarySensor(_WiserDevice):
    """Class representing a Wiser Binary Sensor"""

    @property
    def active(self) -> bool:
        """Get if is active"""
        return self._device_type_data.get("Active")

    @property
    def interacts_with_room_climate(self) -> bool:
        """Get the if interacts with room climate"""
        return self._device_type_data.get("InteractsWithRoomClimate")

    @property
    def type(self) -> str:
        """Get the type of device"""
        return self._device_type_data.get("Type")

    @property
    def enable_notification(self) -> str:
        """Get if notifications is enable"""
        return self._device_type_data.get("EnableNotification")

    @property
    def battery(self) -> _WiserBattery:
        """Get the battery information for the smokealarm"""
        return _WiserBattery(self._data)


class _WiserWindowDoorSensor(_WiserBinarySensor):
    """Class representing a Wiser WindowDoor Sensor"""


class _WiserBinarySensorCollection:
    """Class holding all Wiser Binary Sensors"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserBinarySensor]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, id: int) -> _WiserBinarySensor:
        """
        Gets a binarysensor object from the binary sensor id
        param id: id of binary sensor
        return: _WiserBinarySensor object
        """
        try:
            return [binarysensor for binarysensor in self.all if binarysensor.id == id][
                0
            ]
        except IndexError:
            return None



from datetime import datetime, timedelta

from ..const import TEXT_UNKNOWN


class WiserStatus:
    """Class to hold status object"""

    def __init__(self, data: dict):
        self._data = data if data else {}

    @property
    def uptime(self):
        """Get uptime of hub"""
        sec = timedelta(seconds=int(self._data.get("uptime", 0)))
        d = datetime(1, 1, 1) + sec
        return f"{d.day-1}d {d.hour:02d}:{d.minute:02d}:{d.second:02d}"

    @property
    def free_heap(self):
        """Get memory free heap"""
        return self._data.get("freeHeap", 0)

    @property
    def lowest_free_heap(self):
        """Get lowest memory free heap"""
        return self._data.get("lowestFreeHeap", 0)

    @property
    def last_reset_reason(self):
        """Get last reset reason"""
        return self._data.get("lastResetReason", TEXT_UNKNOWN)

    @property
    def task_usage_enabled(self):
        """Get of task usage enabled"""
        return self._data.get("taskUsageEnabled", False)



import inspect

from . import _LOGGER
from .const import TEXT_UNKNOWN, WISERSYSTEM
from .rest_controller import _WiserRestController


class _WiserAutomation:
    def __init__(
        self, wiser_rest_controller: _WiserRestController, automation_data: dict
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._automation_data = automation_data

    async def _send_command(self, cmd: dict) -> bool:
        """
        Send system control command to Wiser Hub
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        result = await self._wiser_rest_controller._send_command(WISERSYSTEM, cmd)
        if result:
            _LOGGER.debug(
                "Wiser hub - %s command successful", format(inspect.stack()[1].function)
            )
            return True
        return False

    @property
    def id(self) -> int:
        return self._automation_data.get("id", 0)

    @property
    def name(self) -> str:
        return self._automation_data.get("Name", TEXT_UNKNOWN)

    @property
    def enabled(self) -> bool:
        return self._automation_data.get("Enabled", False)

    async def trigger(self):
        """Activate automation"""
        return await self._send_command({"TriggerAutomation": self.id})

    @property
    def notification_enabled(self) -> bool:
        return self._automation_data.get("EnableNotification", False)

    async def enable_notification(self):
        """Activate automation"""
        return await self._send_command({"EnableNotification": self.id})


class _WiserAutomationCollection(object):
    def __init__(
        self, wiser_rest_controller: _WiserRestController, automations_data: dict
    ):
        self._automation_data = automations_data
        self._automations = []
        self._wiser_rest_controller = wiser_rest_controller
        self._build()

    def _build(self):
        for automation in self._automation_data:
            self._automations.append(
                _WiserAutomation(self._wiser_rest_controller, automation)
            )

    @property
    def all(self) -> list[_WiserAutomation]:
        """Return list of automations"""
        return self._automations

    @property
    def count(self) -> int:
        """Count of automations"""
        return len(self._automations)

    def get_by_id(self, automation_id: int) -> _WiserAutomation:
        try:
            return [
                automation for automation in self.all if automation.id == automation_id
            ][0]
        except IndexError:
            return None



import codecs
import os.path

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__VERSION__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="aioWiserHeatAPI",
    version=get_version("aioWiserHeatAPI/__init__.py"),
    author="Mark Parker",
    author_email="msparker@sky.com",
    description="An AsyncIO API for controlling the Drayton Wiser Heating system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/msp1974/aioWiserHeatAPI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "zeroconf>=0.131.0",
        "aiohttp>=3.9.3",
        "aiofiles>=23.2.1",
        "pyyaml>=6.0.1",
    ],
    python_requires=">=3.11",
    entry_points={
        "console_scripts": ["wiser = aioWiserHeatAPI.cli:main"],
    },
)



import inspect
from typing import Union

from . import _LOGGER
from .const import (
    TEXT_OFF,
    TEXT_ON,
    TEXT_UNKNOWN,
    WISERDEVICE,
    WiserDeviceModeEnum,
    WiserLightLedIndicatorEnum,
    WiserLightPowerOnBehaviourEnum,
)
from .helpers.device import _WiserElectricalDevice
from .helpers.misc import is_value_in_list


class _WiserOutputRange(object):
    """Data structure for min/max output range"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def minimum(self) -> int:
        """Get min value"""
        if self._data:
            return self._data.get("Minimum")
        return None

    @property
    def maximum(self) -> int:
        """Get max value"""
        if self._data:
            return self._data.get("Maximum")
        return None


class _WiserLight(_WiserElectricalDevice):
    """Class representing a Wiser Light device"""

    async def _send_command(self, cmd: dict, device_level: bool = False):
        """
        Send control command to the smart plug
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        if device_level:
            result = await self._wiser_rest_controller._send_command(
                WISERDEVICE.format(self.id), cmd
            )
            if result:
                self._data = result
        else:
            result = await self._wiser_rest_controller._send_command(
                self._endpoint.format(self.light_id), cmd
            )
            if result:
                self._device_type_data = result
        if result:
            _LOGGER.debug(
                "Wiser light - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def control_source(self) -> str:
        """Get the current control source of the light"""
        return self._device_type_data.get("ControlSource", TEXT_UNKNOWN)

    @property
    def current_state(self) -> str:
        """Get if light is on"""
        return self._device_type_data.get("CurrentState", 0)

    @property
    def is_dimmable(self) -> bool:
        """Get if the light is dimmable"""
        return True if self._device_type_data.get("IsDimmable", False) else False

    @property
    def is_on(self) -> bool:
        """Get if the light is on"""
        return True if self.current_state == TEXT_ON else False

    @property
    def light_id(self) -> int:
        """Get id of shutter"""
        return self._device_type_data.get("id", 0)

    # added by LGO
    # Hub V2  new features

    @property
    def is_led_indicator_supported(self) -> bool:
        """Get is led indicator supported for the light"""

        return (
            True
            if self._device_type_data.get("IsLedIndicatorSupported", False)
            else False
        )

    @property
    def is_power_on_behaviour_supported(self) -> bool:
        """Get is_power on behaviour supported for the light"""
        return (
            True
            if self._device_type_data.get("IsPowerOnBehaviourSupported", False)
            else False
        )

    @property
    def available_led_indicator(self):
        """Get available led indicator"""
        return [action.value for action in WiserLightLedIndicatorEnum]

    @property
    def led_indicator(self) -> str:
        """Get  led indicator for the light"""
        return self._device_type_data.get("LedIndicator", TEXT_UNKNOWN)

    async def set_led_indicator(
        self, led_indicator: Union[WiserLightLedIndicatorEnum, str]
    ) -> bool:
        if isinstance(led_indicator, WiserLightLedIndicatorEnum):
            led_indicator = led_indicator.value
        if is_value_in_list(led_indicator, self.available_led_indicator):
            return await self._send_command({"LedIndicator": led_indicator})
        else:
            raise ValueError(
                f"{led_indicator} is not a valid mode.  Valid modes are {self.available_led_indicator}"
            )

    @property
    def is_output_mode_supported(self) -> bool:
        """Get is output mode supported for the light"""
        return (
            True
            if self._device_type_data.get("IsOutputModeSupported", False)
            else False
        )

    @property
    def output_mode(self) -> str:
        """Get is output mode supported for the light"""
        return self._device_type_data.get("OutputMode", TEXT_UNKNOWN)

    # end added by LGO

    @property
    def target_state(self) -> int:
        """Get target state of light"""
        return self._device_type_data.get("TargetState", 0)

    async def turn_on(self) -> bool:
        """
        Turn on the light at current brightness level
        return: boolean
        """
        return await self._send_command({"RequestOverride": {"State": TEXT_ON}})

    async def turn_off(self) -> bool:
        """
        Turn off the light
        return: boolean
        """
        return await self._send_command({"RequestOverride": {"State": TEXT_OFF}})


class _WiserDimmableLight(_WiserLight):
    """Class representing a Wiser Dimmable Light device"""

    @property
    def current_level(self) -> int:
        """Get amount light is on"""
        return self._device_type_data.get("CurrentLevel", 0)

    @property
    def current_percentage(self) -> int:
        """Get percentage amount light is on"""
        return self._device_type_data.get("CurrentPercentage", 0)

    async def set_current_percentage(self, percentage: int):
        """Set current brightness percentage"""
        if percentage >= 0 and percentage <= 100:
            return await self._send_command(
                {
                    "RequestOverride": {
                        "State": TEXT_ON,
                        "Percentage": percentage,
                    }
                }
            )
        else:
            raise ValueError("Brightness level percentage must be between 0 and 100")

    @property
    def manual_level(self) -> int:
        """Get manual level of light"""
        return self._device_type_data.get("ManualLevel", 0)

    @property
    def override_level(self) -> int:
        """Get override level of light"""
        return self._device_type_data.get("OverrideLevel", 0)

    @property
    def output_range(self) -> _WiserOutputRange:
        """Get output range min/max."""
        # TODO: Add setter for min max values
        return _WiserOutputRange(self._device_type_data.get("OutputRange", None))

    @property
    def scheduled_percentage(self) -> int:
        """Get the scheduled percentage for the light"""
        return self._data.get("ScheduledPercentage", 0)

    @property
    def target_percentage(self) -> int:
        """Get target percentage brightness of light"""
        return self._device_type_data.get("TargetPercentage", 0)

    # added by LGO
    # Hub V2  new features

    @property
    def available_power_on_behaviour(self):
        """Get available led indicator"""
        return [action.value for action in WiserLightPowerOnBehaviourEnum]

    @property
    def power_on_behaviour(self) -> str:
        """Get power on behaviour for the light"""
        return self._device_type_data.get("PowerOnBehaviour", TEXT_UNKNOWN)

    async def set_power_on_behaviour(
        self, power_on_behaviour: Union[WiserLightPowerOnBehaviourEnum, str]
    ) -> bool:
        if isinstance(power_on_behaviour, WiserLightPowerOnBehaviourEnum):
            power_on_behaviour = power_on_behaviour.value
        if is_value_in_list(power_on_behaviour, self.available_power_on_behaviour):
            return await self._send_command({"PowerOnBehaviour": power_on_behaviour})
        else:
            raise ValueError(
                f"{power_on_behaviour} is not a valid mode.  Valid modes are {self.available_power_on_behaviour}"
            )

    @property
    def power_on_level(self) -> int:
        """Get power on level for the light"""
        return self._device_type_data.get("PowerOnLevel", None)

    # end added by LGO


class _WiserLightCollection(object):
    """Class holding all wiser lights"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserLight]:
        return list(self._items)

    @property
    def available_modes(self) -> list[str]:
        return [mode.value for mode in WiserDeviceModeEnum]

    @property
    def count(self) -> int:
        return len(self.all)

    @property
    def dimmable_lights(self) -> list[_WiserDimmableLight]:
        return list(
            dimmable_lights
            for dimmable_lights in self.all
            if dimmable_lights.is_dimmable
        )

    @property
    def onoff_lights(self) -> list[_WiserLight]:
        return list(
            onoff_lights for onoff_lights in self.all if not onoff_lights.is_dimmable
        )

    def get_by_id(self, light_id: int) -> _WiserLight:
        """
        Gets a Light object from the Lights device id
        param id: device id of shutter
        return: _WiserShutter object
        """
        try:
            return [light for light in self.all if light.id == light_id][0]
        except IndexError:
            return None

    def get_by_light_id(self, light_id: int) -> _WiserLight:
        """
        Gets a Light object from the Lights id
        param id: id of light
        return: _WiserLight object
        """
        try:
            return [light for light in self.all if light.light_id == light_id][0]
        except IndexError:
            return None

    def get_by_room_id(self, room_id: int) -> list[_WiserLight]:
        """
        Gets a Light object from the Lights room_id
        param id: room_id of light
        return: list of _WiserLight objects
        """
        return [light for light in self.all if light.room_id == room_id]



from .const import TEMP_OFF, TEXT_UNKNOWN, WISERHEATINGACTUATOR, WiserTempLimitsEnum
from .helpers.device import _WiserDevice
from .helpers.equipment import _WiserEquipment
from .helpers.temp import _WiserTemperatureFunctions as tf
from .rest_controller import _WiserRestController


class _WiserTemperatureSensor:
    """Data structure for plug in temp sensor"""

    def __init__(
        self, data: dict, wiser_rest_controller: _WiserRestController, actuator_id
    ):
        self._data = data
        self._id = actuator_id
        self._wiser_rest_controller = wiser_rest_controller

    def _send_command(self, cmd: dict):
        """
        Send control command to the temp sensor controller
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        return self._wiser_rest_controller._send_command(
            WISERHEATINGACTUATOR.format(self._id), cmd
        )

    @property
    def measured_temperature(self) -> float:
        """Get the temperature measured by the temperature sensor"""
        return tf._from_wiser_temp(
            self._data.get("MeasuredTemperature", None), "current"
        )

    @property
    def maximum_temperature(self) -> float:
        """Get the maximum temperature setting"""
        return tf._from_wiser_temp(
            self._data.get("MaximumTemperature", None), "floorHeatingMax"
        )

    async def set_maximum_temperature(self, temp: float):
        """Set the maximum temperature setting"""
        if temp >= WiserTempLimitsEnum.floorHeatingMax.value.get(
            "min"
        ) and temp <= WiserTempLimitsEnum.floorHeatingMax.value.get("max"):
            return await self._send_command(
                {
                    "FloorTemperatureSensor": {
                        "MaximumTemperature": tf._to_wiser_temp(temp, "floorHeatingMax")
                    }
                }
            )
        raise ValueError("Max temperature can only be between 5C and 40C")

    @property
    def minimum_temperature(self) -> float:
        """Get the minimum temperature setting"""
        return tf._from_wiser_temp(
            self._data.get("MinimumTemperature", None), "floorHeatingMin"
        )

    async def set_minimum_temperature(self, temp: float):
        """Set the minimum temperature setting"""
        if temp >= WiserTempLimitsEnum.floorHeatingMin.value.get(
            "min"
        ) and temp <= WiserTempLimitsEnum.floorHeatingMin.value.get("max"):
            return await self._send_command(
                {
                    "FloorTemperatureSensor": {
                        "MinimumTemperature": tf._to_wiser_temp(temp, "floorHeatingMin")
                    }
                }
            )
        raise ValueError("Max temperature can only be between 5C and 39C")

    @property
    def sensor_type(self) -> str:
        """Get the sensor type"""
        return self._data.get("SensorType", TEXT_UNKNOWN)

    @property
    def status(self) -> str:
        """Get the status"""
        return self._data.get("Status", TEXT_UNKNOWN)

    @property
    def temperature_offset(self) -> float:
        """Get the temperature offset"""
        return tf._from_wiser_temp(self._data.get("Offset", None), "floorHeatingOffset")

    async def set_temperature_offset(self, temp: float):
        """Set the temperature offset"""
        if temp >= WiserTempLimitsEnum.floorHeatingOffset.value.get(
            "min"
        ) and temp <= WiserTempLimitsEnum.floorHeatingOffset.value.get("max"):
            return await self._send_command(
                {
                    "FloorTemperatureSensor": {
                        "Offset": tf._to_wiser_temp(temp, "floorHeatingOffset")
                    }
                }
            )
        raise ValueError("Offset temperature can only be between -9C and 9C")


class _WiserHeatingActuator(_WiserDevice):
    """Class representing a Wiser Heating Actuator device"""

    @property
    def current_target_temperature(self) -> float:
        """Get the current target temperature setting"""
        return tf._from_wiser_temp(
            self._device_type_data.get("OccupiedHeatingSetPoint", TEMP_OFF)
        )

    @property
    def current_temperature(self) -> float:
        """Get the current measured temperature"""
        return tf._from_wiser_temp(
            self._device_type_data.get("MeasuredTemperature", TEMP_OFF),
            "current",
        )

    @property
    def delivered_power(self) -> int:
        """Get the amount of power delivered over time"""
        return self._device_type_data.get("CurrentSummationDelivered", None)

    @property
    def equipment_id(self) -> int:
        """Get equipment id (v2 hub)"""
        return self._device_type_data.get("EquipmentId", 0)

    @property
    def equipment(self) -> _WiserEquipment | None:
        """Get equipment data"""
        return (
            _WiserEquipment(self._device_type_data.get("EquipmentData"))
            if self._device_type_data.get("EquipmentData")
            else None
        )

    @property
    def floor_temperature_sensor(self) -> _WiserTemperatureSensor:
        """Get the temperature sensor object"""
        if self._device_type_data.get("FloorTemperatureSensor"):
            return _WiserTemperatureSensor(
                self._device_type_data.get("FloorTemperatureSensor", {}),
                self._wiser_rest_controller,
                self.id,
            )

    @property
    def instantaneous_power(self) -> int:
        """Get the amount of current passing through the device now"""
        return self._device_type_data.get("InstantaneousDemand", None)

    @property
    def output_type(self) -> str:
        """Get output type"""
        return self._device_type_data.get("OutputType", TEXT_UNKNOWN)


class _WiserHeatingActuatorCollection(object):
    """Class holding all wiser heating actuators"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserHeatingActuator]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, id: int) -> _WiserHeatingActuator:
        """
        Gets a Heating Actuator object from the Heating Actuators id
        param id: id of smart valve
        return: _WiserSmartValve object
        """
        try:
            return [
                heating_actuator
                for heating_actuator in self.all
                if heating_actuator.id == id
            ][0]
        except IndexError:
            return None
